#!/usr/bin/env python3
# _*_ coding:utf-8 _*_

# SPDX-FileCopyrightText: 2023 UnionTech Software Technology Co., Ltd.

# SPDX-License-Identifier: GPL-2.0-only
import os

import letmego

from setting.globalconfig import GlobalConfig
from src import log

@letmego.mark
@log
class CmdPublicWidget:

    def reboot(self):
        """reboot"""
        os.system(f"echo '{GlobalConfig.PASSWORD}' | sudo -S reboot")


#!/usr/bin/env python3
# _*_ coding:utf-8 _*_

# SPDX-FileCopyrightText: 2023 UnionTech Software Technology Co., Ltd.

# SPDX-License-Identifier: GPL-2.0-only
# pylint: disable=C0114
# pylint: disable=C0301,C0116,W0613,W1514,C0103,C0103
import json
import os
import re
from os.path import exists
from urllib.parse import urlencode

from setting.globalconfig import GlobalConfig
from src  import logger
from src.pms._base import _Base
from src.pms._base import runs_id_cmd_log


class Send2Pms(_Base):
    """发送数据到pms"""

    __author__ = "huangmingqiang@uniontech.com"

    def post_to_pms(self, run_case_id=None, case_id=None, result=None, **kwargs):
        base_url = "https://pms.uniontech.com/testtask-runCase"
        # 每次请求时，获取steps_id
        run_case_html_url = f"{base_url}-{run_case_id}.html"
        res = self.rx.open_url(run_case_html_url)
        if res == "":
            # 测试套件回填时地址不一样
            run_case_html_url = f"{base_url}-0-{run_case_id}-1.html"
            res = self.rx.open_url(run_case_html_url)
        steps_id = re.findall(r"name='steps\[(\d+)\]'", res)
        if steps_id and steps_id[0]:
            steps_id = steps_id[0]
        else:
            return 201
        # 构造post数据
        data = {
            f"steps[{steps_id}]": result,
            f"reals[{steps_id}]": "",
            "case": case_id,
            "version": "1",
            f"labels{steps_id}[]": "",
            f"files{steps_id}[]": "",
        }
        bytes_data = urlencode(data).encode("utf-8")
        # post请求接口
        res = self.rx.session.open(
            fullurl=run_case_html_url, data=bytes_data, timeout=10
        )
        status_code = res.status
        return status_code

    def send2pms(self, case_res_path, data_send_result_csv):
        """发送数据到PMS"""
        if exists(case_res_path):
            for case_name_json in os.listdir(case_res_path):
                if not case_name_json.endswith(".json"):
                    continue
                # 读取本地json文件中的数据
                with open(f"{case_res_path}/{case_name_json}", "r") as f:
                    data = json.load(f)
                if not exists(data_send_result_csv):
                    with open(data_send_result_csv, "w+") as f:
                        pass
                with open(data_send_result_csv, "r") as f:
                    reqeusted = f.read()
                case_name = case_name_json.split(".")[0]

                # 如果这条用例已经回填过数据
                if case_name in reqeusted:
                    continue

                with open(data_send_result_csv, "a+") as f:
                    for _ in range(int(GlobalConfig.SEND_PMS_RETRY_NUMBER)):
                        # 请求
                        status_code = self.post_to_pms(**data)
                        if status_code == 200:
                            logger.info(f"{runs_id_cmd_log(data)} 数据回填成功 😃")
                            f.write(f"{case_name},request_ok\n")
                            break
                    else:
                        logger.info(f"{runs_id_cmd_log(data)} 数据回填失败 😡")
                        f.write(f"{case_name},request_fail\n")

    @staticmethod
    def case_res_path(taskid):
        """case_res_path"""
        return f"{GlobalConfig.REPORT_PATH}/pms_{taskid}"

    @classmethod
    def data_send_result_csv(cls, taskid):
        """data_send_result_csv"""
        return f"{cls.case_res_path(taskid)}/send_pms_{taskid}.csv"



#!/usr/bin/env python3
# _*_ coding:utf-8 _*_

# SPDX-FileCopyrightText: 2023 UnionTech Software Technology Co., Ltd.

# SPDX-License-Identifier: GPL-2.0-only
# pylint: disable=C0114
from setting.globalconfig import GlobalConfig


def skipif_platform(args: str):
    """平台跳过
    skipif_platform-aarch64
    """
    _skip_key = args.split("&")
    for key in _skip_key:
        if GlobalConfig.SYS_ARCH == key:
            return True
    return False


def skipif_xdg_type(args: str):
    """skipif wayland or x11
    skipif_xdg_type-wayland
    """
    _skip_key = args.split("&")
    for key in _skip_key:
        if GlobalConfig.DISPLAY_SERVER == key:
            return True
    return False


def skipif_cpu_name(args: str):
    """skipif cpu name
    使用 sudo dmidecode -s system-product-name 查看机器的cpu型号
    剔除中横线和&符号，比如：KLVV-W5821，标签记录为 KLVVW5821
    """
    import os
    _skip_key = args.split("&")
    for key in _skip_key:
        if os.popen(
                f"echo '{GlobalConfig.PASSWORD}'| "
                "sudo -S dmidecode -s system-product-name | awk '{print $NF}'"
        ).read().strip("\n").replace("-", "").replace("&", "") == key:
            return True
    return False



# -*- coding: utf-8 -*-

# SPDX-FileCopyrightText: 2023 UnionTech Software Technology Co., Ltd.

# SPDX-License-Identifier: GPL-2.0-only
# pylint: disable=C0114
from os.path import join
from os.path import dirname
from os.path import abspath

from setting.globalconfig import _GlobalConfig
from setting.globalconfig import GetCfg


# pylint: disable=too-few-public-methods
class _Config(_GlobalConfig):
    """
    Application library configuration
    """

    # pylint: disable=too-few-public-methods
    class DirName:
        """DirName"""

        # pylint: disable=too-few-public-methods
        class SubDirName:
            """SubDirName"""
            UI_INI = "ui.ini"
            PIC_RES = "pic_res"

    ABSOLUTE_BASE_PATH = dirname(abspath(__file__))
    # res 目录下子目录绝对路径
    UI_INI_PATH = join(ABSOLUTE_BASE_PATH, DirName.SubDirName.UI_INI)
    PIC_RES_PATH = join(ABSOLUTE_BASE_PATH, DirName.SubDirName.PIC_RES)

    # config ini object
    CONFIG_FILE_PATH = join(ABSOLUTE_BASE_PATH, "config.ini")
    cfg = GetCfg(CONFIG_FILE_PATH, "config")

    # config.ini
    DESKTOP_CONFIG = cfg.get("DESKTOP_CONFIG")
    FTP_ADDRESS = cfg.get("FTP_ADDRESS")


Config = _Config()



#!/usr/bin/env python3
# _*_ coding:utf-8 _*_

# SPDX-FileCopyrightText: 2023 UnionTech Software Technology Co., Ltd.

# SPDX-License-Identifier: GPL-2.0-only
import sys
import os
from configparser import ConfigParser
from configparser import NoSectionError


def cli():
    """Command Line Create Project"""
    args = sys.argv[1:]
    project_name = "youqu"
    root_dir = os.path.dirname(os.path.abspath(__file__))
    if len(args) == 1:
        project_name = args[0]
    elif len(args) > 1:
        raise ValueError("只能接受 1 个参数")

    if os.path.exists(f"./{project_name}"):
        raise FileExistsError(f"目录:{project_name} 已经存在，请确认您要创建的项目名称！")
    os.system(f"rm -rf /tmp/{project_name}")
    os.makedirs(f"/tmp/{project_name}")
    os.system(f"cp -r {root_dir}/. /tmp/{project_name}/")
    os.system(f"mv /tmp/{project_name} .")
    os.system(f"rm -rf {project_name}/startproject.py")
    os.system(f"rm -rf {project_name}/.gitignore")
    for root, dirs, files in os.walk(f"./{project_name}"):
        for d in dirs:
            if d == "__pycache__":
                os.system(f"rm -rf {root}/{d}")
    conf = ConfigParser()
    try:
        conf.read(f"{root_dir}/CURRENT")
        youqu_version = conf.get("current", "tag")
    except NoSectionError:
        youqu_version = None
    print(f"The project: [\033[0;32m{project_name}\033[0m],has been created by youqu{f'-{youqu_version}' if youqu_version else ''}")


if __name__ == '__main__':
    cli()



#!/usr/bin/env python3
# _*_ coding:utf-8 _*_

# SPDX-FileCopyrightText: 2023 UnionTech Software Technology Co., Ltd.

# SPDX-License-Identifier: GPL-2.0-only
# pylint: disable=C0114
import letmego

from src import logger, log
from src import Src
from public.open_mode_public_widget.config import Config

@letmego.mark
@log
# pylint: disable=too-many-ancestors
class OpenModePublicWidget(Src):
    """
    文件打开方式封装
    """

    @classmethod
    def click_open_mode_image(cls, pic):
        """
         点击右键菜单中的相应选项
        :param pic: 图片路径
        :return
        """
        cls.click(*cls.find_image(f"{Config.PIC_RES_PATH}/" + pic))

    @classmethod
    def click_txt_editor_in_open_mode_window_by_image(cls):
        """
         点击打开方式窗口中的“文本编辑器”
        :return:
        """
        cls.click_open_mode_image("text_editor")
        logger.info("点击打开方式窗口中的“文本编辑器")

    @classmethod
    def click_file_manager_in_open_mode_window_by_image(cls):
        """
         点击打开方式窗口中的“文件管理器”
        :return:
        """
        cls.click_open_mode_image("file_manager")
        logger.info("点击打开方式窗口中的“文件管理器”")

    @classmethod
    def click_determine_in_open_mode_window_by_image(cls):
        """
         点击打开方式窗口中的“确定”
        :return:
        """
        cls.click_open_mode_image("determine")
        logger.info("点击确认")



# -*- coding: utf-8 -*-
# This file was modified by UnionTech Software Technology Co., Ltd. in 2022/08/11
# pylint: disable=C0301,W0511,C0103,C0116,W0107,R1705,W0613,R0913,W0612,W0703,C0201,C0121,C0114,W0603,C0123
__author__ = "zhaoyouzhi@uniontech.com"
__contributor__ = "huangmingqiang@uniontech.com"

import os
import sys
from functools import wraps
from time import sleep

# from PIL import Image

from src.cmdctl import CmdCtl

if sys.version_info[0] == 2 or sys.version_info[0:2] in ((3, 1), (3, 7)):
    # Python 2 and 3.1 and 3.2 uses collections.Sequence
    import collections

# In seconds. Any duration less than this is rounded to 0.0 to instantly move
# the mouse.
MINIMUM_DURATION = 0.1
# If sleep_amount is less than MINIMUM_DURATION, sleep() will be a no-op and the mouse cursor moves there instantly.
# TODO: This value should vary with the platform. http://stackoverflow.com/q/1133857
MINIMUM_SLEEP = 0.05
STEP_SLEEP = 10

# The number of seconds to pause after EVERY public function call. Useful for debugging:
PAUSE = 0.1  # Tenth-second pause by default.

FAILSAFE = True

Point = collections.namedtuple("Point", "x y")
Size = collections.namedtuple("Size", "width height")

bshift = False

LEFT = "left"
MIDDLE = "middle"
RIGHT = "right"
PRIMARY = "left"
SECONDARY = "secondary"

dbus_cmd = "dbus-send --session --dest=com.deepin.Autotool --print-reply  /com/deepin/Autotool com.deepin.Autotool"

def context_manager(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        tool_status = os.popen("ps -aux |  grep wayland_autotool | grep -v grep").read()
        if not tool_status:
            os.environ["QT_WAYLAND_SHELL_INTEGRATION"] = "kwayland-shell"
            os.environ["XDG_SESSION_DESKTOP"] = "Wayland"
            os.environ["XDG_SESSION_TYPE"] = "wayland"
            os.environ["WAYLAND_DISPLAY"] = "wayland-0"
            os.environ["GDMSESSION"] = "Wayland"
            os.environ["DBUS_SESSION_BUS_ADDRESS"] = "unix:path=/run/user/1000/bus"
            res = os.popen("nohup wayland_autotool > result.log 2>&1 &").read()
            if "未找到命令" in res:
                raise EnvironmentError("wayland_autotool没有安装")
            sleep(2)

        return func(*args, **kwargs)

    return wrapper


@context_manager
def system(cmd):
    CmdCtl.run_cmd(cmd, out_debug_flag=False, command_log=False)


@context_manager
def popen(cmd):
    return CmdCtl.run_cmd(cmd, out_debug_flag=False, command_log=False)


KEY_NAMES = {
    ' ': 57,
    '\'': 40,
    '*': 55,
    '+': 78,
    ',': 51,
    '-': 12,
    '.': 52,
    '/': 53,
    '0': 11,
    '1': 2,
    '2': 3,
    '4': 5,
    '5': 6,
    '3': 4,
    '6': 7,
    '7': 8,
    '8': 9,
    '9': 10,
    ';': 39,
    '=': 13,
    '[': 26,
    '\\': 43,
    ']': 27,
    '`': 41,
    'a': 30,
    'b': 48,
    'c': 46,
    'd': 32,
    'e': 18,
    'f': 33,
    'g': 34,
    'h': 35,
    'i': 23,
    'j': 36,
    'k': 37,
    'l': 38,
    'm': 50,
    'n': 49,
    'o': 24,
    'p': 25,
    'q': 16,
    'r': 19,
    's': 31,
    't': 20,
    'u': 22,
    'v': 47,
    'w': 17,
    'x': 45,
    'y': 21,
    'z': 44,
    'add': 78,
    'alt': 56,
    'altleft': 56,
    'altright': 100,
    'backspace': 14,
    'capslock': 58,
    'ctrl': 29,
    'ctrlleft': 29,
    'ctrlright': 97,
    'del': 111,
    'delete': 111,
    'down': 108,
    'end': 107,
    'enter': 28,
    'esc': 1,
    'escape': 1,
    'f1': 59,
    'f10': 68,
    'f11': 87,
    'f12': 88,
    'f13': 183,
    'f14': 184,
    'f15': 185,
    'f16': 186,
    'f17': 187,
    'f18': 188,
    'f19': 189,
    'f2': 60,
    'f20': 190,
    'f21': 191,
    'f22': 192,
    'f23': 193,
    'f24': 194,
    'f3': 61,
    'f4': 62,
    'f5': 63,
    'f6': 64,
    'f7': 65,
    'f8': 66,
    'f9': 67,
    'home': 172,
    'insert': 110,
    'left': 105,
    'num0': 82,
    'num1': 79,
    'num2': 80,
    'num3': 81,
    'num4': 75,
    'num5': 76,
    'num6': 77,
    'num7': 71,
    'num8': 72,
    'num9': 73,
    'numlock': 69,
    'pagedown': 109,
    'pageup': 104,
    'pgdn': 109,
    'pgup': 104,
    'print': 210,
    'right': 106,
    'scrolllock': 70,
    'shift': 42,
    'shiftleft': 42,
    'shiftright': 54,
    'space': 57,
    'tab': 15,
    'up': 103,
    'volumedown': 114,
    'volumeup': 115,
    'win': 125,
    'winleft': 125,
    'winright': 126,
    ")": 11,
    "!": 2,
    "@": 3,
    "#": 4,
    "$": 5,
    "%": 6,
    "^": 7,
    "&": 8,
    "(": 10,
    "_": 12,
    "~": 41,
    "{": 26,
    "}": 27,
    "|": 43,
    ":": 39,
    "\"": 40,
    "<": 51,
    ">": 52,
    "?": 53,
    'A': 30,
    'B': 48,
    'C': 46,
    'D': 32,
    'E': 18,
    'F': 33,
    'G': 34,
    'H': 35,
    'I': 23,
    'J': 36,
    'K': 37,
    'L': 38,
    'M': 50,
    'N': 49,
    'O': 24,
    'P': 25,
    'Q': 16,
    'R': 19,
    'S': 31,
    'T': 20,
    'U': 22,
    'V': 47,
    'W': 17,
    'X': 45,
    'Y': 21,
    'Z': 44,
    '。': 52
}

KEY_NAMES_S = {
    ")": 11,
    "!": 2,
    "@": 3,
    "#": 4,
    "$": 5,
    "%": 6,
    "^": 7,
    "&": 8,
    "(": 10,
    "_": 12,
    "~": 41,
    "{": 26,
    "}": 27,
    "|": 43,
    ":": 39,
    "\"": 40,
    "<": 51,
    ">": 52,
    "?": 53,
    'A': 30,
    'B': 48,
    'C': 46,
    'D': 32,
    'E': 18,
    'F': 33,
    'G': 34,
    'H': 35,
    'I': 23,
    'J': 36,
    'K': 37,
    'L': 38,
    'M': 50,
    'N': 49,
    'O': 24,
    'P': 25,
    'Q': 16,
    'R': 19,
    'S': 31,
    'T': 20,
    'U': 22,
    'V': 47,
    'W': 17,
    'X': 45,
    'Y': 21,
    'Z': 44
}


class PyAutoGUIException(Exception):
    """
    PyAutoGUI code will raise this exception class for any invalid actions. If PyAutoGUI raises some other exception,
    you should assume that this is caused by a bug in PyAutoGUI itself. (Including a failure to catch potential
    exceptions raised by PyAutoGUI.)
    """
    pass


def getPointOnLine(x1, y1, x2, y2, n):
    """
    Returns an (x, y) tuple of the point that has progressed a proportion ``n`` along the line defined by the two
    ``x1``, ``y1`` and ``x2``, ``y2`` coordinates.

    This function was copied from pytweening module, so that it can be called even if PyTweening is not installed.
    """
    x = ((x2 - x1) * n) + x1
    y = ((y2 - y1) * n) + y1
    return (x, y)


def linear(n):
    """
    Returns ``n``, where ``n`` is the float argument between ``0.0`` and ``1.0``. This function is for the default
    linear tween for mouse moving functions.

    This function was copied from PyTweening module, so that it can be called even if PyTweening is not installed.
    """

    # We use this function instead of pytweening.linear for the default tween function just in case pytweening couldn't be imported.
    if not 0.0 <= n <= 1.0:
        raise PyAutoGUIException("Argument must be between 0.0 and 1.0.")
    return n


def position(x=None, y=None):
    """
    Returns the current xy coordinates of the mouse cursor as a two-integer tuple.
    Args:
      x (int, None, optional) - If not None, this argument overrides the x in
        the return value.
      y (int, None, optional) - If not None, this argument overrides the y in
        the return value.
    Returns:
      (x, y) tuple of the current xy coordinates of the mouse cursor.
    NOTE: The position() function doesn't check for failsafe.
    """
    result = popen(
        "dbus-send --session --dest=com.deepin.Autotool --print-reply  /com/deepin/Autotool com.deepin.Autotool.getPos | grep double | awk \'{printf \"%d\\n\",$2}\'"
    ).split("\n")
    posx = int(result[0].strip())
    posy = int(result[1].strip())

    if x is not None:  # If set, the x parameter overrides the return value.
        posx = int(x)
    if y is not None:  # If set, the y parameter overrides the return value.
        posy = int(y)
    return Point(posx, posy)


def _normalizeXYArgs(firstArg, secondArg):
    """
    Returns a ``Point`` object based on ``firstArg`` and ``secondArg``, which are the first two arguments passed to
    several PyAutoGUI functions. If ``firstArg`` and ``secondArg`` are both ``None``, returns the current mouse cursor
    position.

    ``firstArg`` and ``secondArg`` can be integers, a sequence of integers, or a string representing an image filename
    to find on the screen (and return the center coordinates of).
    """
    if firstArg is None and secondArg is None:
        return position()
    else:
        return Point(firstArg, secondArg)


def size():
    result = popen(
        "dbus-send --session --dest=com.deepin.Autotool --print-reply  /com/deepin/Autotool com.deepin.Autotool.getSize | grep int32 | awk \'{printf \"%d\\n\",$2}\'"
    ).split("\n")
    posx = int(result[0].strip())
    posy = int(result[1].strip())
    if 0 == posx or 0 == posy:
        return size()
    return Size(posx, posy)


def moveTo(x=None, y=None, duration=0.0, tween=linear, logScreenshot=False, _pause=True):
    startx, starty = position()

    x = int(x) if x is not None else startx
    y = int(y) if y is not None else starty

    width, height = size()

    # Make sure x and y are within the screen bounds.
    # x = max(0, min(x, width - 1))
    # y = max(0, min(y, height - 1))

    # If the duration is small enough, just move the cursor there instantly.
    steps = [(x, y)]

    if duration > MINIMUM_DURATION:
        # Non-instant moving/dragging involves tweening:
        num_steps = max(width, height)
        sleep_amount = duration / num_steps
        if sleep_amount < MINIMUM_SLEEP:
            num_steps = int(duration / MINIMUM_SLEEP)
            sleep_amount = duration / num_steps
        # print(str(startx) + "," + str(starty))
        # print(str(x) + "," + str(y))
        steps = [getPointOnLine(startx, starty, x, y, tween(n / num_steps)) for n in range(num_steps)]
        # Making sure the last position is the actual destination.
        # print(steps)
        steps.append((x, y))
    # print(steps)
    for tweenX, tweenY in steps:
        if len(steps) > 1:
            # A single step does not require tweening.
            sleep(sleep_amount + 0.01)

        tweenX = int(round(tweenX))
        tweenY = int(round(tweenY))
        system(
            f"{dbus_cmd}.moveTo int32:{str(tweenX)} int32:{str(tweenY)}"
        )


def mouseDown(x=None, y=None, button=PRIMARY, duration=0.0, tween=linear, logScreenshot=None, _pause=True):
    x, y = _normalizeXYArgs(x, y)
    moveTo(x, y, duration, tween, logScreenshot, _pause)
    system(
        f"{dbus_cmd}.mouseDown string:{button}"
    )
    sleep(0.3)


def mouseUp(x=None, y=None, button=PRIMARY, duration=0.0, tween=linear, logScreenshot=None, _pause=True):
    x, y = _normalizeXYArgs(x, y)
    moveTo(x, y, duration, tween, logScreenshot, _pause)
    system(
        f"{dbus_cmd}.mouseUp string:{button}"
    )


def click(
        x=None, y=None, clicks=1, interval=0.0, button=PRIMARY, duration=0.0, tween=linear, logScreenshot=None,
        _pause=True
):
    x, y = _normalizeXYArgs(x, y)
    moveTo(x, y, duration, tween, logScreenshot, _pause)
    for i in range(clicks):
        if button in (LEFT, MIDDLE, RIGHT):
            system(
                f"{dbus_cmd}.click string:{button}"
            )
        sleep(interval)


def leftClick(x=None, y=None, interval=0.0, duration=0.0, tween=linear, logScreenshot=None, _pause=True):
    click(x, y, 1, interval, LEFT, duration, tween, logScreenshot, _pause=_pause)


def rightClick(x=None, y=None, interval=0.0, duration=0.0, tween=linear, logScreenshot=None, _pause=True):
    click(x, y, 1, interval, RIGHT, duration, tween, logScreenshot, _pause=_pause)


def middleClick(x=None, y=None, interval=0.0, duration=0.0, tween=linear, logScreenshot=None, _pause=True):
    click(x, y, 1, interval, MIDDLE, duration, tween, logScreenshot, _pause=_pause)


def doubleClick(x=None, y=None, interval=0.0, button=LEFT, duration=0.0, tween=linear, logScreenshot=None, _pause=True):
    click(x, y, 2, interval, button, duration, tween, logScreenshot, _pause=False)


def tripleClick(x=None, y=None, interval=0.0, button=LEFT, duration=0.0, tween=linear, logScreenshot=None, _pause=True):
    click(x, y, 3, interval, button, duration, tween, logScreenshot, _pause=False)


def scroll(clicks, x=None, y=None, logScreenshot=None, _pause=True):
    if type(x) in (tuple, list):
        x, y = x[0], x[1]
    x, y = position(x, y)
    moveTo(x, y)
    vscroll(clicks, x, y, logScreenshot, _pause)


def hscroll(clicks, x=None, y=None, logScreenshot=None, _pause=True):
    if type(x) in (tuple, list):
        x, y = x[0], x[1]
    x, y = position(x, y)
    moveTo(x, y)
    step = clicks / STEP_SLEEP
    for i in range(0, STEP_SLEEP):
        system(
            f"{dbus_cmd}.hscroll double:{str(-step)}"
        )
        sleep(0.05)


def vscroll(clicks, x=None, y=None, logScreenshot=None, _pause=True):
    if type(x) in (tuple, list):
        x, y = x[0], x[1]
    x, y = position(x, y)
    moveTo(x, y)
    step = clicks / STEP_SLEEP
    for i in range(0, STEP_SLEEP):
        system(
            f"{dbus_cmd}.vscroll double:{str(-step)}"
        )
        sleep(0.05)


def dragTo(
        x=None, y=None, duration=0.0, tween=linear, button=PRIMARY, logScreenshot=None, _pause=True, mouseDownUp=True
):
    x, y = _normalizeXYArgs(x, y)
    if mouseDownUp:
        mouseDown(button=button, logScreenshot=False, _pause=False)
    moveTo(x, y, duration, tween)
    if mouseDownUp:
        mouseUp(button=button, logScreenshot=False, _pause=False)


def dragRel(
        xOffset=0, yOffset=0, duration=0.0, tween=linear, button=PRIMARY, logScreenshot=None, _pause=True,
        mouseDownUp=True
):
    if xOffset is None:
        xOffset = 0
    if yOffset is None:
        yOffset = 0

    if type(xOffset) in (tuple, list):
        xOffset, yOffset = xOffset[0], xOffset[1]

    if xOffset == 0 and yOffset == 0:
        return  # no-op case

    mousex, mousey = position()
    mousex = mousex + xOffset
    mousey = mousey + mousey
    if mouseDownUp:
        mouseDown(button=button, logScreenshot=False, _pause=False)
    moveTo(mousex, mousey, duration, tween)
    if mouseDownUp:
        mouseUp(button=button, logScreenshot=False, _pause=False)


def keyDown(key, logScreenshot=None, _pause=True):
    global bshift
    try:
        if len(key) > 1:
            key = key.lower()
        if key in KEY_NAMES_S.keys():
            system(f"{dbus_cmd}.keyDown uint32:{str(KEY_NAMES['shiftleft'])}")
            bshift = True
        system(f"{dbus_cmd}.keyDown uint32:{str(KEY_NAMES[key])}")
    except Exception:
        if bshift == True:
            system(f"{dbus_cmd}.keyUp uint32:{str(KEY_NAMES['shiftleft'])}")
            bshift = False
        pass


def keyUp(key, logScreenshot=None, _pause=True):
    global bshift
    try:
        if len(key) > 1:
            key = key.lower()
        system(
            f"{dbus_cmd}.keyUp uint32:{str(KEY_NAMES[key])}"
        )
        if key in KEY_NAMES_S.keys():
            system(f"{dbus_cmd}.keyUp uint32:{str(KEY_NAMES['shiftleft'])}")
            bshift = False
    except Exception:
        if bshift == True:
            system(f"{dbus_cmd}.keyUp uint32:{str(KEY_NAMES['shiftleft'])}")
            bshift = False
        pass


def press(keys, presses=1, interval=0.0, logScreenshot=None, _pause=True):
    if type(keys) == str:
        if len(keys) > 1:
            keys = keys.lower()
        keys = [keys]  # If keys is 'enter', convert it to ['enter'].
    else:
        lowerKeys = []
        for s in keys:
            if len(s) > 1:
                lowerKeys.append(s.lower())
            else:
                lowerKeys.append(s)
        keys = lowerKeys
    interval = float(interval)
    for i in range(presses):
        for k in keys:
            keyDown(k)
            keyUp(k)
        sleep(interval)


def typewrite(message, interval=0.0, logScreenshot=None, _pause=True):
    interval = float(interval)  # TODO - this should be taken out.

    for c in message:
        if len(c) > 1:
            c = c.lower()
        press(c, _pause=False)
        sleep(interval)


write = typewrite  # In PyAutoGUI 1.0, write() replaces typewrite().


def hotkey(*args, **kwargs):
    """Performs key down presses on the arguments passed in order, then performs
    key releases in reverse order.

    The effect is that calling hotkey('ctrl', 'shift', 'c') would perform a
    "Ctrl-Shift-C" hotkey/keyboard shortcut press.

    Args:
      key(s) (str): The series of keys to press, in order. This can also be a
        list of key strings to press.
      interval (float, optional): The number of seconds in between each press.
        0.0 by default, for no pause in between presses.

    Returns:
      None
    """
    interval = float(kwargs.get("interval", 0.0))  # TODO - this should be taken out.

    for c in args:
        if len(c) > 1:
            c = c.lower()
        # print("keydown:" + c)
        keyDown(c)
        sleep(interval)
    for c in reversed(args):
        if len(c) > 1:
            c = c.lower()
        # print("keyup:" + c)
        keyUp(c)
        sleep(interval)


def screenshot():
    command = "dbus-send --session --print-reply --dest=org.kde.KWin /Screenshot org.kde.kwin.Screenshot.screenshotFullscreen"
    image_path = popen(command).strip().split("\n")[1].split("\"")[1].strip()
    # if image_path != "":
    #     image = Image.open(image_path)
    #     return image
    # else:
    #     return None



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
from .config import config
from . import path
from . import predicate
from . import rawinput
from .logging import debugLogger as logger
from .utils import doDelay, Blinker, Lock

from time import sleep
from types import LambdaType
import gi
from gi.repository import GLib
import os
import sys

try:
    import pyatspi
    import Accessibility
except ImportError:  # pragma: no cover
    raise ImportError("Error importing the AT-SPI bindings")

"""Makes some sense of the AT-SPI API

The tree API handles various things for you:
    - fixes most timing issues
    - can automatically generate (hopefully) highly-readable logs of what the
script is doing
    - traps various UI malfunctions, raising exceptions for them (again,
hopefully improving the logs)

The most important class is Node. Each Node is an element of the desktop UI.
There is a tree of nodes, starting at 'root', with applications as its
children, with the top-level windows and dialogs as their children. The various
widgets that make up the UI appear as descendents in this tree. All of these
elements (root, the applications, the windows, and the widgets) are represented
as instances of Node in a tree (provided that the program of interest is
correctly exporting its user-interface to the accessibility system). The Node
class is a mixin for Accessible and the various Accessible interfaces.

The Action class represents an action that the accessibility layer exports as
performable on a specific node, such as clicking on it. It's a wrapper around
Accessibility.Action.

We often want to look for a node, based on some criteria, and this is provided
by the Predicate class.

Dogtail implements a high-level searching system, for finding a node (or
nodes) satisfying whatever criteria you are interested in. It does this with
a 'backoff and retry' algorithm. This fixes most timing problems e.g. when a
dialog is in the process of opening but hasn't yet done so.

If a search fails, it waits 'config.searchBackoffDuration' seconds, and then
tries again, repeatedly. After several failed attempts (determined by
config.searchWarningThreshold) it will start sending warnings about the search
to the debug log. If it still can't succeed after 'config.searchCutoffCount'
attempts, it raises an exception containing details of the search. You can see
all of this process in the debug log by setting 'config.debugSearching' to True

We also automatically add a short delay after each action
('config.defaultDelay' gives the time in seconds). We'd hoped that the search
backoff and retry code would eliminate the need for this, but unfortunately we
still run into timing issues. For example, Evolution (and probably most
other apps) set things up on new dialogs and wizard pages as they appear, and
we can run into 'setting wars' where the app resets the widgetry to defaults
after our script has already filled out the desired values, and so we lose our
values. So we give the app time to set the widgetry up before the rest of the
script runs.

The classes trap various UI malfunctions and raise exceptions that better
describe what went wrong. For example, they detects attempts to click on an
insensitive UI element and raise a specific exception for this.

Unfortunately, some applications do not set up the 'sensitive' state
correctly on their buttons (e.g. Epiphany on form buttons in a web page). The
current workaround for this is to set config.ensureSensitivity=False, which
disables the sensitivity testing.
"""
__author__ = """Zack Cerza <zcerza@redhat.com>,
David Malcolm <dmalcolm@redhat.com>
"""

if config.checkForA11y:
    from .utils import checkForA11y
    checkForA11y()

# We optionally import the bindings for libWnck.
try:
    gi.require_version('Wnck', '3.0')
    from gi.repository import Wnck
    gotWnck = True  # pragma: no cover
except (ImportError, ValueError):
    # Skip this warning, since the functionality is almost entirely nonworking anyway.
    # print "Warning: Dogtail could not import the Python bindings for
    # libwnck. Window-manager manipulation will not be available."
    gotWnck = False

haveWarnedAboutChildrenLimit = False


class SearchError(Exception):
    pass


class NotSensitiveError(Exception):
    """
    The widget is not sensitive.
    """
    message = "Cannot %s %s. It is not sensitive."

    def __init__(self, action):
        self.action = action

    def __str__(self):
        return self.message % (self.action.name, self.action.node.getLogString())


class ActionNotSupported(Exception):
    """
    The widget does not support the requested action.
    """
    message = "Cannot do '%s' action on %s"

    def __init__(self, actionName, node):
        self.actionName = actionName
        self.node = node

    def __str__(self):
        return self.message % (self.actionName, self.node.getLogString())


class Action(object):
    """
    Class representing an action that can be performed on a specific node
    """
    # Valid types of actions we know about. Feel free to add any you see.
    types = ('click',
             'press',
             'release',
             'activate',
             'jump',
             'check',
             'dock',
             'undock',
             'open',
             'menu')

    def __init__(self, node, action, index):
        self.node = node
        self.__action = action
        self.__index = index

    @property
    def name(self):
        return self.__action.getName(self.__index)

    @property
    def description(self):
        return self.__action.getDescription(self.__index)

    @property
    def keyBinding(self):
        return self.__action.getKeyBinding(self.__index)

    def __str__(self):
        return "[action | %s | %s ]" % \
            (self.name, self.keyBinding)

    def do(self):
        """
        Performs the given tree.Action, with appropriate delays and logging.
        """
        logger.log(str("%s on %s") % (str(self.name), self.node.getLogString()))
        if not self.node.sensitive:
            if config.ensureSensitivity:
                raise NotSensitiveError(self)
            else:
                nSE = NotSensitiveError(self)
                logger.log("Warning: " + str(nSE))
        if config.blinkOnActions:
            self.node.blink()
        result = self.__action.doAction(self.__index)
        doDelay(config.actionDelay)
        return result


class Node(object):
    """
    A node in the tree of UI elements. This class is mixed in with
    Accessibility.Accessible to both make it easier to use and to add
    additional functionality. It also has a debugName which is set up
    automatically when doing searches.
    """

    def __setupUserData(self):
        try:
            len(self.user_data)
        except (AttributeError, TypeError):
            self.user_data = {}

    @property
    def debugName(self):
        """debug name assigned during search operations"""
        self.__setupUserData()
        return self.user_data.get('debugName', None)

    @debugName.setter
    def debugName(self, debugName):
        self.__setupUserData()
        self.user_data['debugName'] = debugName

    #
    # Accessible
    #
    @property
    def dead(self):
        """
        Is the node dead (defunct)?
        """
        try:
            if self.roleName == 'invalid':
                return True
            self.role
            self.name
            if len(self) > 0:
                self[0]
        except:
            return True
        return False

    @property
    def children(self):
        """
        A list of this Accessible's children
        """
        if self.parent and self.parent.roleName == 'hyper link':
            print(self.parent.role)
            return []
        children = []
        childCount = self.childCount
        if childCount > config.childrenLimit:
            global haveWarnedAboutChildrenLimit
            if not haveWarnedAboutChildrenLimit:
                logger.log("Only returning %s children. You may change "
                           "config.childrenLimit if you wish. This message will only"
                           " be printed once." % str(config.childrenLimit))
                haveWarnedAboutChildrenLimit = True
                childCount = config.childrenLimit
        for i in range(childCount):
            # Workaround for GNOME bug #465103
            # also solution for GNOME bug #321273
            try:
                child = self[i]
            except LookupError:
                child = None
            if child:
                children.append(child)

        invalidChildren = childCount - len(children)
        if invalidChildren and config.debugSearching:
            logger.log(str("Skipped %s invalid children of %s") %
                       (invalidChildren, str(self)))
        try:
            ht = self.queryHypertext()
            for li in range(ht.getNLinks()):
                link = ht.getLink(li)
                for ai in range(link.nAnchors):
                    child = link.getObject(ai)
                    if child == self:
                        continue
                    child.__setupUserData()
                    child.user_data['linkAnchor'] = \
                        LinkAnchor(node=child,
                                   hypertext=ht,
                                   linkIndex=li,
                                   anchorIndex=ai)
                    children.append(child)
        except (NotImplementedError, AttributeError):
            pass

        return children

    roleName = property(Accessibility.Accessible.getRoleName)

    role = property(Accessibility.Accessible.getRole)

    indexInParent = property(Accessibility.Accessible.getIndexInParent)

    #
    # Action
    #

    # Needed to be renamed from doAction due to conflicts
    # with 'Accessibility.Accessible.doAction' in gtk3 branch
    def doActionNamed(self, name):
        """
        Perform the action with the specified name. For a list of actions
        supported by this instance, check the 'actions' property.
        """
        actions = self.actions
        if name in actions:
            return actions[name].do()
        raise ActionNotSupported(name, self)

    @property
    def actions(self):
        """
        A dictionary of supported action names as keys, with Action objects as
        values. Common action names include:

        'click' 'press' 'release' 'activate' 'jump' 'check' 'dock' 'undock'
        'open' 'menu'
        """
        actions = {}
        try:
            action = self.queryAction()
            for i in range(action.nActions):
                a = Action(self, action, i)
                actions[action.getName(i)] = a
        finally:
            return actions

    @property
    def combovalue(self):
        """
        The value (as a string) currently selected in the combo box.
        """
        return self.name

    @combovalue.setter
    def combovalue(self, value):
        logger.log(str("Setting combobox %s to '%s'") % (self.getLogString(), str(value)))
        self.childNamed(childName=value).doActionNamed('click')
        doDelay()

    #
    # Hypertext and Hyperlink
    #

    @property
    def URI(self):
        try:
            return self.user_data['linkAnchor'].URI
        except (KeyError, AttributeError):
            raise NotImplementedError

    #
    # Text and EditableText
    #

    @property
    def text(self):
        """
        For instances with an AccessibleText interface, the text as a
        string. This is read-only, unless the instance also has an
        AccessibleEditableText interface. In this case, you can write values
        to the attribute. This will get logged in the debug log, and a delay
        will be added.

        If this instance corresponds to a password entry, use the passwordText
        property instead.
        """

        try:
            return self.queryText().getText(0, -1)
        except NotImplementedError:
            return None

    @text.setter
    def text(self, text):
        try:
            if config.debugSearching:
                msg = "Setting text of %s to %s"
                # Let's not get too crazy if 'text' is really large...
                # FIXME: Sometimes the next line screws up Unicode strings.
                if len(text) > 140:
                    txt = text[:134] + " [...]"
                else:
                    txt = text
                logger.log(str(msg) % (self.getLogString(), str("'%s'") % str(txt)))
            self.queryEditableText().setTextContents(text)
        except NotImplementedError:
            raise AttributeError("can't set attribute")

    @property
    def caretOffset(self):
        """
        For instances with an AccessibleText interface, the caret offset as an integer.
        """
        return self.queryText().caretOffset

    @caretOffset.setter
    def caretOffset(self, offset):
        return self.queryText().setCaretOffset(offset)

    #
    # Component
    #

    @property
    def position(self):
        """
        A tuple containing the position of the Accessible: (x, y)
        """
        return self.queryComponent().getPosition(pyatspi.DESKTOP_COORDS)

    @property
    def size(self):
        """
        A tuple containing the size of the Accessible: (w, h)
        """
        return self.queryComponent().getSize()

    @property
    def extents(self):
        """
        A tuple containing the location and size of the Accessible: (x, y, w, h)
        """
        try:
            ex = self.queryComponent().getExtents(pyatspi.DESKTOP_COORDS)
            return (ex.x, ex.y, ex.width, ex.height)
        except NotImplementedError:
            return None

    def contains(self, x, y):
        try:
            return self.queryComponent().contains(x, y, pyatspi.DESKTOP_COORDS)
        except NotImplementedError:
            return False

    def getChildAtPoint(self, x, y):
        node = self
        while True:
            try:
                child = node.queryComponent().getAccessibleAtPoint(x, y, pyatspi.DESKTOP_COORDS)
                if child and child.contains(x, y):
                    node = child
                else:
                    break
            except NotImplementedError:
                break
        if node and node.contains(x, y):
            return node
        else:
            return None

    def grabFocus(self):
        """
        Attempts to set the keyboard focus to this Accessible.
        """
        return self.queryComponent().grabFocus()

    def click(self, button=1):
        """
        Generates a raw mouse click event, using the specified button.
            - 1 is left,
            - 2 is middle,
            - 3 is right.
        """
        logger.log(str("Clicking on %s") % self.getLogString())
        clickX = self.position[0] + self.size[0] / 2
        clickY = self.position[1] + self.size[1] / 2
        if config.debugSearching:
            logger.log(str("raw click on %s %s at (%s,%s)") %
                       (str(self.name), self.getLogString(), str(clickX), str(clickY)))
        rawinput.click(clickX, clickY, button)

    def doubleClick(self, button=1):
        """
        Generates a raw mouse double-click event, using the specified button.
        """
        clickX = self.position[0] + self.size[0] / 2
        clickY = self.position[1] + self.size[1] / 2
        if config.debugSearching:
            logger.log(str("raw click on %s %s at (%s,%s)") %
                       (str(self.name), self.getLogString(), str(clickX), str(clickY)))
        rawinput.doubleClick(clickX, clickY, button)

    def point(self, mouseDelay=None):
        """
        Move mouse cursor to the center of the widget.
        """
        pointX = self.position[0] + self.size[0] / 2
        pointY = self.position[1] + self.size[1] / 2
        logger.log(str("Pointing on %s %s at (%s,%s)") %
                   (str(self.name), self.getLogString(), str(pointX), str(pointY)))
        if config.IS_X11:
            rawinput.registry.generateMouseEvent(pointX, pointY, 'abs')
        else:
            from . import rawinput_wayland
            rawinput_wayland.moveTo(pointX, pointY)
        if mouseDelay:
            doDelay(mouseDelay)
        else:
            doDelay()

    #
    # RelationSet
    #
    @property
    def labeler(self):
        """
        'labeller' (read-only list of Node instances):
        The node(s) that is/are a label for this node. Generated from 'relations'.
        """
        relationSet = self.getRelationSet()
        for relation in relationSet:
            if relation.getRelationType() == pyatspi.RELATION_LABELLED_BY:
                if relation.getNTargets() == 1:
                    return relation.getTarget(0)
                targets = []
                for i in range(relation.getNTargets()):
                    targets.append(relation.getTarget(i))
                return targets
    labeller = labeler

    @property
    def labelee(self):
        """
        'labellee' (read-only list of Node instances):
        The node(s) that this node is a label for. Generated from 'relations'.
        """
        relationSet = self.getRelationSet()
        for relation in relationSet:
            if relation.getRelationType() == pyatspi.RELATION_LABEL_FOR:
                if relation.getNTargets() == 1:
                    return relation.getTarget(0)
                targets = []
                for i in range(relation.getNTargets()):
                    targets.append(relation.getTarget(i))
                return targets
    labellee = labelee

    #
    # StateSet
    #
    @property
    def sensitive(self):
        """
        Is the Accessible sensitive (i.e. not greyed out)?
        """
        return self.getState().contains(pyatspi.STATE_SENSITIVE)

    @property
    def showing(self):
        """
        Is the Accessible really showing (rendered and visible) on the screen?
        """
        return self.getState().contains(pyatspi.STATE_SHOWING)

    @property
    def focusable(self):
        """
        Is the Accessible capable of having keyboard focus?
        """
        return self.getState().contains(pyatspi.STATE_FOCUSABLE)

    @property
    def focused(self):
        """
        Does the Accessible have keyboard focus?
        """
        return self.getState().contains(pyatspi.STATE_FOCUSED)

    @property
    def checked(self):
        """
        Is the Accessible a checked checkbox?
        """
        return self.getState().contains(pyatspi.STATE_CHECKED)

    @property
    def isChecked(self):
        """
        Is the Accessible a checked checkbox? Compatibility property, same as Node.checked.
        """
        return self.checked

    @property
    def visible(self):
        """
        Is the Accessible set to be visible? A widget with set attribute
        'visible' is supposed to be shown and doesn't need to be actually
        rendered. On the other hand, a widget with unset attribute 'visible'
        """
        return self.getState().contains(pyatspi.STATE_VISIBLE)

    #
    # Selection
    #

    def selectAll(self):
        """
        Selects all children.
        """
        result = self.querySelection().selectAll()
        doDelay()
        return result

    def deselectAll(self):
        """
        Deselects all selected children.
        """
        result = self.querySelection().clearSelection()
        doDelay()
        return result

    def select(self):
        """
        Selects the Accessible.
        """
        try:
            parent = self.parent
        except AttributeError:
            raise NotImplementedError
        result = parent.querySelection().selectChild(self.indexInParent)
        doDelay()
        return result

    def deselect(self):
        """
        Deselects the Accessible.
        """
        try:
            parent = self.parent
        except AttributeError:
            raise NotImplementedError
        result = parent.querySelection().deselectChild(self.indexInParent)
        doDelay()
        return result

    @property
    def isSelected(self):
        """
        Is the Accessible selected? Compatibility property, same as Node.selected.
        """
        try:
            parent = self.parent
        except AttributeError:
            raise NotImplementedError
        return parent.querySelection().isChildSelected(self.indexInParent)

    @property
    def selected(self):
        """
        Is the Accessible selected?
        """
        return self.isSelected

    @property
    def selectedChildren(self):
        """
        Returns a list of children that are selected.
        """
        # TODO: hideChildren for Hyperlinks?
        selection = self.querySelection()
        selectedChildren = []
        for i in range(selection.nSelectedChildren):
            selectedChildren.append(selection.getSelectedChild(i))

    #
    # Value
    #

    @property
    def value(self):
        """
        The value contained by the AccessibleValue interface.
        """
        try:
            return self.queryValue().currentValue
        except NotImplementedError:
            pass

    @value.setter
    def value(self, value):
        """
        Setter for the value contained by the AccessibleValue interface.
        """
        self.queryValue().currentValue = value

    @property
    def minValue(self):
        """
        The minimum value of self.value
        """
        try:
            return self.queryValue().minimumValue
        except NotImplementedError:
            pass

    @property
    def minValueIncrement(self):
        """
        The minimum value increment of self.value
        """
        try:
            return self.queryValue().minimumIncrement
        except NotImplementedError:
            pass

    @property
    def maxValue(self):
        """
        The maximum value of self.value
        """
        try:
            return self.queryValue().maximumValue
        except NotImplementedError:
            pass

    def typeText(self, string):
        """
        Type the given text into the node, with appropriate delays and logging.
        """
        logger.log(str("Typing text into %s: '%s'") % (self.getLogString(), str(string)))

        if self.focusable:
            if not self.focused:
                try:
                    self.grabFocus()
                except Exception:
                    logger.log("Node is focusable but I can't grabFocus!")
            rawinput.typeText(string)
        else:
            logger.log("Node is not focusable; falling back to inserting text")
            et = self.queryEditableText()
            et.insertText(self.caretOffset, string, len(string))
            self.caretOffset += len(string)
            doDelay()

    def keyCombo(self, comboString):
        if config.debugSearching:
            logger.log(str("Pressing keys '%s' into %s") %
                       (str(comboString), self.getLogString()))
        if self.focusable:
            if not self.focused:
                try:
                    self.grabFocus()
                except Exception:
                    logger.log("Node is focusable but I can't grabFocus!")
        else:
            logger.log("Node is not focusable; trying key combo anyway")
        rawinput.keyCombo(comboString)

    def getLogString(self):
        """
        Get a string describing this node for the logs,
        respecting the config.absoluteNodePaths boolean.
        """
        if config.absoluteNodePaths:
            return self.getAbsoluteSearchPath()
        else:
            return str(self)

    def satisfies(self, pred):
        """
        Does this node satisfy the given predicate?
        """
        # the logic is handled by the predicate:
        assert isinstance(pred, predicate.Predicate)
        return pred.satisfiedByNode(self)

    def dump(self, type='plain', fileName=None):
        from . import dump
        dumper = getattr(dump, type)
        dumper(self, fileName)

    def getAbsoluteSearchPath(self):
        """
        FIXME: this needs rewriting...
        Generate a SearchPath instance giving the 'best'
        way to find the Accessible wrapped by this node again, starting
        at the root and applying each search in turn.

        This is somewhat analagous to an absolute path in a filesystem,
        except that some of searches may be recursive, rather than just
        searching direct children.

        Used by the recording framework for identifying nodes in a
        persistent way, independent of the style of script being
        written.

        FIXME: try to ensure uniqueness
        FIXME: need some heuristics to get 'good' searches, whatever
        that means
        """
        if config.debugSearchPaths:
            logger.log("getAbsoluteSearchPath(%s)" % self)

        if self.roleName == 'application':
            result = path.SearchPath()
            result.append(predicate.IsAnApplicationNamed(self.name), False)
            return result
        else:
            if self.parent:
                (ancestor, pred, isRecursive) = self.getRelativeSearch()
                if config.debugSearchPaths:
                    logger.log("got ancestor: %s" % ancestor)

                ancestorPath = ancestor.getAbsoluteSearchPath()
                ancestorPath.append(pred, isRecursive)
                return ancestorPath
            else:
                # This should be the root node:
                return path.SearchPath()

    def getRelativeSearch(self):
        """
        Get a (ancestorNode, predicate, isRecursive) triple that identifies the
        best way to find this Node uniquely.
        FIXME: or None if no such search exists?
        FIXME: may need to make this more robust
        FIXME: should this be private?
        """
        if config.debugSearchPaths:
            logger.log("getRelativeSearchPath(%s)" % self)

        assert self
        assert self.parent

        isRecursive = False
        ancestor = self.parent

        # iterate up ancestors until you reach an identifiable one,
        # setting the search to be isRecursive if need be:
        while not self.__nodeIsIdentifiable(ancestor):
            ancestor = ancestor.parent
            isRecursive = True

        # Pick the most appropriate predicate for finding this node:
        if self.labellee:
            if self.labellee.name:
                return (ancestor, predicate.IsLabelledAs(self.labellee.name), isRecursive)

        if self.roleName == 'menu':
            return (ancestor, predicate.IsAMenuNamed(self.name), isRecursive)
        elif self.roleName == 'menu item' or self.roleName == 'check menu item':
            return (ancestor, predicate.IsAMenuItemNamed(self.name), isRecursive)
        elif self.roleName == 'text':
            return (ancestor, predicate.IsATextEntryNamed(self.name), isRecursive)
        elif self.roleName == 'push button':
            return (ancestor, predicate.IsAButtonNamed(self.name), isRecursive)
        elif self.roleName == 'frame':
            return (ancestor, predicate.IsAWindowNamed(self.name), isRecursive)
        elif self.roleName == 'dialog':
            return (ancestor, predicate.IsADialogNamed(self.name), isRecursive)
        else:
            pred = predicate.GenericPredicate(
                name=self.name, roleName=self.roleName)
            return (ancestor, pred, isRecursive)

    def __nodeIsIdentifiable(self, ancestor):
        if ancestor.labellee:
            return True
        elif ancestor.name:
            return True
        elif not ancestor.parent:
            return True
        else:
            return False

    def _fastFindChild(self, pred, recursive=True, showingOnly=None):
        """
        Searches for an Accessible using methods from pyatspi.utils
        """
        if isinstance(pred, predicate.Predicate):
            pred = pred.satisfiedByNode
        if showingOnly is None:
            showingOnly = config.searchShowingOnly
        if showingOnly:
            orig_pred = pred
            pred = lambda n: orig_pred(n) and \
                             n.getState().contains(pyatspi.STATE_SHOWING)
        if not recursive:
            cIter = iter(self)
            while True:
                try:
                    child = next(cIter)
                except StopIteration:
                    break
                if child is not None and pred(child):
                    return child
        else:
            return pyatspi.utils.findDescendant(self, pred)

    def findChild(self, pred, recursive=True, debugName=None, retry=True, requireResult=True, showingOnly=None):
        """
        Search for a node satisyfing the predicate, returning a Node.

        If retry is True (the default), it makes multiple attempts,
        backing off and retrying on failure, and eventually raises a
        descriptive exception if the search fails.

        If retry is False, it gives up after one attempt.

        If requireResult is True (the default), an exception is raised after all
        attempts have failed. If it is false, the function simply returns None.
        """
        def describeSearch(parent, pred, recursive, debugName):
            """
            Internal helper function
            """
            if recursive:
                noun = "descendent"
            else:
                noun = "child"
            if debugName is None:
                debugName = pred.describeSearchResult()
            return str("%s of %s: %s") % (str(noun), parent.getLogString(), str(debugName))

        compare_func = None
        if isinstance(pred, LambdaType):
            compare_func = pred
            if debugName is None:
                debugName = "child satisyfing a custom lambda function"
        else:
            assert isinstance(pred, predicate.Predicate)
            compare_func = pred.satisfiedByNode

        numAttempts = 0
        while numAttempts < config.searchCutoffCount:
            if numAttempts >= config.searchWarningThreshold or config.debugSearching:
                logger.log(str("searching for %s (attempt %i)") %
                           (describeSearch(self, pred, recursive, debugName), numAttempts))

            result = self._fastFindChild(compare_func, recursive, showingOnly=showingOnly)
            if result:
                assert isinstance(result, Node)
                if debugName:
                    result.debugName = debugName
                else:
                    result.debugName = pred.describeSearchResult()
                return result
            else:
                if not retry:
                    break
                numAttempts += 1
                if config.debugSearching or config.debugSleep:
                    logger.log("sleeping for %f" % config.searchBackoffDuration)
                sleep(config.searchBackoffDuration)
        if requireResult:
            raise SearchError(describeSearch(self, pred, recursive, debugName))

    # The canonical "search for multiple" method:
    def findChildren(self, pred, recursive=True, isLambda=False, showingOnly=None):
        """
        Find all children/descendents satisfying the predicate.
        You can also use lambdas in place of pred that will enable search also against
        pure dogtail Node properties (like showing). I.e: "lambda x: x.roleName == 'menu item'
        and x.showing is True". isLambda does not have to be set, it's kept only for api compatibility.
        """
        # always use lambda search, but we keep isLambda param for api compatibility
        compare_func = None
        if isLambda is True or isinstance(pred, LambdaType):
            compare_func = pred
        else:
            assert isinstance(pred, predicate.Predicate)
            compare_func = pred.satisfiedByNode
        if showingOnly is None:
            showingOnly = config.searchShowingOnly
        if showingOnly:
            orig_compare_func = compare_func
            compare_func = lambda n: orig_compare_func(n) and \
                                     n.getState().contains(pyatspi.STATE_SHOWING)

        results = []
        numAttempts = 0
        while numAttempts < config.searchCutoffCount:
            if numAttempts >= config.searchWarningThreshold or config.debugSearching:
                logger.log("a11y errors caught, making attempt %i" % numAttempts)
            try:
                if recursive:
                    results = pyatspi.utils.findAllDescendants(self, compare_func)
                else:
                    results = list(filter(compare_func, self.children))
                break
            except (GLib.GError, TypeError):
                numAttempts += 1
                if numAttempts == config.searchCutoffCount:
                    logger.log("warning: errors caught from the a11y tree, giving up search")
                else:
                    sleep(config.searchBackoffDuration)
                continue
        return results

    # The canonical "search above this node" method:
    def findAncestor(self, pred, showingOnly=None):
        """
        Search up the ancestry of this node, returning the first Node
        satisfying the predicate, or None.
        """
        assert isinstance(pred, predicate.Predicate)
        candidate = self.parent
        while candidate is not None:
            if candidate.satisfies(pred):
                return candidate
            else:
                candidate = candidate.parent
        # Not found:
        return None

    # Various wrapper/helper search methods:
    def child(self, name='', roleName='', description='', label='', recursive=True, retry=True, debugName=None, showingOnly=None):
        """
        Finds a child satisying the given criteria.

        This is implemented using findChild, and hence will automatically retry
        if no such child is found, and will eventually raise an exception. It
        also logs the search.
        """
        return self.findChild(predicate.GenericPredicate(name=name, roleName=roleName, description=description,
                              label=label), recursive=recursive, retry=retry, debugName=debugName, showingOnly=showingOnly)

    def isChild(self, name='', roleName='', description='', label='', recursive=True, retry=False, debugName=None, showingOnly=None):
        """
        Determines whether a child satisying the given criteria exists.

        This is implemented using findChild, but will not automatically retry
        if no such child is found. To make the function retry multiple times set retry to True.
        Returns a boolean value depending on whether the child was eventually found. Similar to
        'child', yet it catches SearchError exception to provide for False results, will raise
        any other exceptions. It also logs the search.
        """
        found = True
        try:
            self.findChild(
                predicate.GenericPredicate(
                    name=name, roleName=roleName, description=description, label=label),
                recursive=recursive, retry=retry, debugName=debugName, showingOnly=showingOnly)
        except SearchError:
            found = False
        return found

    def menu(self, menuName, recursive=True, showingOnly=None):
        """
        Search below this node for a menu with the given name.

        This is implemented using findChild, and hence will automatically retry
        if no such child is found, and will eventually raise an exception. It
        also logs the search.
        """
        return self.findChild(predicate.IsAMenuNamed(menuName=menuName), recursive, showingOnly=showingOnly)

    def menuItem(self, menuItemName, recursive=True, showingOnly=None):
        """
        Search below this node for a menu item with the given name.

        This is implemented using findChild, and hence will automatically retry
        if no such child is found, and will eventually raise an exception. It
        also logs the search.
        """
        return self.findChild(predicate.IsAMenuItemNamed(menuItemName=menuItemName), recursive, showingOnly=showingOnly)

    def textentry(self, textEntryName, recursive=True, showingOnly=None):
        """
        Search below this node for a text entry with the given name.

        This is implemented using findChild, and hence will automatically retry
        if no such child is found, and will eventually raise an exception. It
        also logs the search.
        """
        return self.findChild(predicate.IsATextEntryNamed(textEntryName=textEntryName), recursive, showingOnly=showingOnly)

    def button(self, buttonName, recursive=True, showingOnly=None):
        """
        Search below this node for a button with the given name.

        This is implemented using findChild, and hence will automatically retry
        if no such child is found, and will eventually raise an exception. It
        also logs the search.
        """
        return self.findChild(predicate.IsAButtonNamed(buttonName=buttonName), recursive, showingOnly=showingOnly)

    def childLabelled(self, labelText, recursive=True, showingOnly=None):
        """
        Search below this node for a child labelled with the given text.

        This is implemented using findChild, and hence will automatically retry
        if no such child is found, and will eventually raise an exception. It
        also logs the search.
        """
        return self.findChild(predicate.IsLabelledAs(labelText), recursive, showingOnly=showingOnly)

    def childNamed(self, childName, recursive=True, showingOnly=None):
        """
        Search below this node for a child with the given name.

        This is implemented using findChild, and hence will automatically retry
        if no such child is found, and will eventually raise an exception. It
        also logs the search.
        """
        return self.findChild(predicate.IsNamed(childName), recursive, showingOnly=showingOnly)

    def tab(self, tabName, recursive=True, showingOnly=None):
        """
        Search below this node for a tab with the given name.

        This is implemented using findChild, and hence will automatically retry
        if no such child is found, and will eventually raise an exception. It
        also logs the search.
        """
        return self.findChild(predicate.IsATabNamed(tabName=tabName), recursive, showingOnly=showingOnly)

    def getUserVisibleStrings(self):
        """
        Get all user-visible strings in this node and its descendents.

        (Could be implemented as an attribute)
        """
        result = []
        if self.name:
            result.append(self.name)
        if self.description:
            result.append(self.description)
        try:
            children = self.children
        except Exception:
            return result
        for child in children:
            result.extend(child.getUserVisibleStrings())
        return result

    def blink(self):
        """
        Blink, baby!
        """
        if not self.extents:
            return False
        else:
            (x, y, w, h) = self.extents
            Blinker(x, y, w, h)
            return True


class LinkAnchor(object):
    """
    Class storing info about an anchor within an Accessibility.Hyperlink, which
    is in turn stored within an Accessibility.Hypertext.
    """

    def __init__(self, node, hypertext, linkIndex, anchorIndex):
        self.node = node
        self.hypertext = hypertext
        self.linkIndex = linkIndex
        self.anchorIndex = anchorIndex

    @property
    def link(self):
        return self.hypertext.getLink(self.linkIndex)

    @property
    def URI(self):
        return self.link.getURI(self.anchorIndex)


class Root (Node):
    """
    FIXME:
    """

    def applications(self):
        """
        Get all applications.
        """
        return root.findChildren(predicate.GenericPredicate(roleName="application"), recursive=False, showingOnly=False)

    def application(self, appName, retry=True):
        """
        Gets an application by name, returning an Application instance
        or raising an exception.

        This is implemented using findChild, and hence will automatically retry
        if no such child is found, and will eventually raise an exception. It
        also logs the search.
        """
        return root.findChild(predicate.IsAnApplicationNamed(appName), recursive=False, retry=retry, showingOnly=False)


class Application (Node):
    def dialog(self, dialogName, recursive=False, showingOnly=None):
        """
        Search below this node for a dialog with the given name,
        returning a Window instance.

        This is implemented using findChild, and hence will automatically retry
        if no such child is found, and will eventually raise an exception. It
        also logs the search.

        FIXME: should this method activate the dialog?
        """
        return self.findChild(predicate.IsADialogNamed(dialogName=dialogName), recursive, showingOnly=showingOnly)

    def window(self, windowName, recursive=False, showingOnly=None):
        """
        Search below this node for a window with the given name,
        returning a Window instance.

        This is implemented using findChild, and hence will automatically retry
        if no such child is found, and will eventually raise an exception. It
        also logs the search.

        FIXME: this bit isn't true:
        The window will be automatically activated (raised and focused
        by the window manager) if wnck bindings are available.
        """
        result = self.findChild(predicate.IsAWindowNamed(windowName=windowName), recursive, showingOnly=showingOnly)
        # FIXME: activate the WnckWindow ?
        # if gotWnck:
        #       result.activate()
        return result

    def getWnckApplication(self, showingOnly=None):  # pragma: no cover
        """
        Get the wnck.Application instance for this application, or None

        Currently implemented via a hack: requires the app to have a
        window, and looks up the application of that window

        wnck.Application can give you the pid, the icon, etc

        FIXME: untested
        """
        window = self.child(roleName='frame', showingOnly=showingOnly)
        if window:
            wnckWindow = window.getWnckWindow()
            return wnckWindow.get_application()


class Window (Node):

    def getWnckWindow(self):  # pragma: no cover
        """
        Get the wnck.Window instance for this window, or None
        """
        # FIXME: this probably needs rewriting:
        screen = Wnck.screen_get_default()

        # You have to force an update before any of the wnck methods
        # do anything:
        screen.force_update()

        for wnckWindow in screen.get_windows():
            # FIXME: a dubious hack: search by window title:
            if wnckWindow.get_name() == self.name:
                return wnckWindow

    def activate(self):  # pragma: no cover
        """
        Activates the wnck.Window associated with this Window.

        FIXME: doesn't yet work
        """
        wnckWindow = self.getWnckWindow()
        # Activate it with a timestamp of 0; this may confuse
        # alt-tabbing through windows etc:
        # FIXME: is there a better way of getting a timestamp?
        # gdk_x11_get_server_time (), with a dummy window
        wnckWindow.activate(0)


class Wizard (Window):
    """
    Note that the buttons of a GnomeDruid were not accessible until
    recent versions of libgnomeui.  This is
    http://bugzilla.gnome.org/show_bug.cgi?id=157936
    and is fixed in gnome-2.10 and gnome-2.12 (in CVS libgnomeui);
    there's a patch attached to that bug.

    This bug is known to affect FC3; fixed in FC5
    """

    def __init__(self, node, debugName=None):
        Node.__init__(self, node)
        if debugName:
            self.debugName = debugName
        logger.log(str("%s is on '%s' page") % (self, str(self.getPageTitle())))

    def currentPage(self):
        """
        Get the current page of this wizard

        FIXME: this is currently a hack, supporting only GnomeDruid
        """
        pageHolder = self.child(roleName='panel')
        for child in pageHolder.children:
            if child.showing:
                return child
        raise "Unable to determine current page of %s" % self

    def getPageTitle(self):
        """
        Get the string title of the current page of this wizard

        FIXME: this is currently a total hack, supporting only GnomeDruid
        """
        currentPage = self.currentPage()
        return currentPage.child(roleName='panel').child(roleName='panel').child(roleName='label', recursive=False).text

    def clickForward(self):
        """
        Click on the 'Forward' button to advance to next page of wizard.

        It will log the title of the new page that is reached.

        FIXME: what if it's Next rather than Forward ???

        This will only work if your libgnomeui has accessible buttons;
        see above.
        """
        fwd = self.child("Forward")
        fwd.click()

        # Log the new wizard page; it's helpful when debugging scripts
        logger.log(str("%s is now on '%s' page") % (self, str(self.getPageTitle())))
        # FIXME disabled for now (can't get valid page titles)

    def clickApply(self):
        """
        Click on the 'Apply' button to advance to next page of wizard.
        FIXME: what if it's Finish rather than Apply ???

        This will only work if your libgnomeui has accessible buttons;
        see above.
        """
        fwd = self.child("Apply")
        fwd.click()

        # FIXME: debug logging?

Accessibility.Accessible.__bases__ = (
    Application, Root, Node,) + Accessibility.Accessible.__bases__

try:
    root = pyatspi.Registry.getDesktop(0)
    root.debugName = 'root'
except Exception:  # pragma: no cover
    # Warn if AT-SPI's desktop object doesn't show up.
    logger.log("Error: AT-SPI's desktop is not visible. Do you have accessibility enabled?")

# Check that there are applications running. Warn if none are.
children = root.children
if not children:  # pragma: no cover
    logger.log(
        "Warning: AT-SPI's desktop is visible but it has no children. Are you running any AT-SPI-aware applications?")
del children

# sniff also imports from tree and we don't want to run this code from sniff itself
if not os.path.exists('/tmp/sniff_running.lock'):
    if not os.path.exists('/tmp/sniff_refresh.lock'):  # may have already been locked by dogtail.procedural
        # 'tell' newly opened sniff not to use auto-refresh while script using this module is running
        sniff_lock = Lock(lockname='sniff_refresh.lock', randomize=False, unlockOnExit=True)
        try:
            sniff_lock.lock()
        except OSError:  # pragma: no cover
            pass
# elif 'sniff' not in sys.argv[0]:
#     print("Dogtail: Warning: Running sniff has been detected.")
#     print("Please make sure sniff has the 'Auto Refresh' disabled.")
#     print("NOTE: Running scripts with sniff present is not recommended.")


# Convenient place to set some debug variables:
# config.debugSearching = True
# config.absoluteNodePaths = True
# config.logDebugToFile = False



#!/usr/bin/env python3
# _*_ coding:utf-8 _*_

# SPDX-FileCopyrightText: 2023 UnionTech Software Technology Co., Ltd.

# SPDX-License-Identifier: GPL-2.0-only
# pylint: disable=C0114
# pylint: disable=E0401,C0413,R0902,R0913,R0914,W0613,C0301,C0415,C0103
import re
import sys
from configparser import ConfigParser
from itertools import cycle
from os import listdir
from os import makedirs
from os import popen
from os import system
from os.path import exists
from os.path import splitext
# from threading import Thread
# from multiprocessing import Process as Thread
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import wait
from concurrent.futures import ALL_COMPLETED
from time import sleep
from time import strftime

from allure_custom import AllureCustom
from allure_custom.conf import setting

from setting.globalconfig import GlobalConfig

setting.html_title = GlobalConfig.REPORT_TITLE
setting.report_name = GlobalConfig.REPORT_NAME
setting.report_language = GlobalConfig.REPORT_LANGUAGE

from src.cmdctl import CmdCtl
from src import logger
from src.rtk._base import Args
from src.rtk._base import transform_app_name


class RemoteRunner:
    """
    远程执行器：控制多台测试机远程执行用例。
    在 setting/remote.ini 里面配置要执行的测试机信息。
    """

    __author__ = "Mikigo <huangmingqiang@uniontech.com>"

    def __init__(
            self,
            remote_kwargs: dict = None,
            local_kwargs: dict = None,
    ):
        self.remote_kwargs = remote_kwargs
        self.local_kwargs = local_kwargs
        logger("INFO")
        conf = ConfigParser()
        conf.read(f"{GlobalConfig.SETTING_PATH}/remote.ini")
        self.parallel = conf.getboolean("remote", "PARALLEL")
        self.clean_server_report_dir = conf.getboolean(
            "remote", "CLEAN_SERVER_REPORT_DIR"
        )
        self.clean_client_report_dir = conf.getboolean(
            "remote", "CLEAN_CLIENT_REPORT_DIR"
        )
        self.send_code = conf.getboolean("remote", "SEND_CODE")
        self.scan = conf.getint("remote", "SCAN")
        self.client_env = conf.getboolean("remote", "BUILD_ENV")
        self.client_password = conf.get("remote", "CLIENT_PASSWORD")

        self._default = {
            Args.client_password.value: remote_kwargs.get("client_password") or self.client_password,
        }

        self.ini_client_dict = {
            op: [
                conf.get(op, "user"),
                conf.get(op, "ip"),
                conf.get(op, "password", fallback=self._default.get(Args.client_password.value)),
            ]
            for op in filter(lambda x: "client" in x, conf.sections())
        }

        cli_client_dict = {}
        if remote_kwargs.get("clients"):
            clients = remote_kwargs.get("clients").split("/")
            for index, client in enumerate(clients):
                client_info = re.findall(r"^(.+?)@(\d+\.\d+\.\d+\.\d+):{0,1}(.*?)$", client)
                if client_info:
                    _c = list(client_info[0])
                    if _c[2] == "":
                        _c[2] = self._default.get(Args.client_password.value)
                    cli_client_dict[f"client{index + 1}"] = _c

        if not cli_client_dict and not self.ini_client_dict:
            raise ValueError(
                "未获取到测试机信息,请检查 setting/remote.ini 中 CLIENT LIST 是否配置，"
                "或通过命令行 remote -c user@ip:password 传入。"
            )

        self.default = {
            Args.app_name.value: transform_app_name(local_kwargs.get("app_name") or GlobalConfig.APP_NAME),
            Args.clients.value: cli_client_dict or self.ini_client_dict,
            Args.send_code.value: remote_kwargs.get("send_code") or self.send_code,
            Args.build_env.value: remote_kwargs.get("build_env") or self.client_env,
            Args.parallel.value: remote_kwargs.get("parallel") or self.parallel,
        }
        # 客户端地址
        if "/home/" not in GlobalConfig.ROOT_DIR:
            raise EnvironmentError
        self.server_project_path = "/".join(GlobalConfig.ROOT_DIR.split("/")[3:])
        self.client_report_path = (
            lambda x: f"/home/{x}/{self.server_project_path}/report"
        )
        self.client_allure_report_path = lambda \
                x: f"/home/{x}/{self.server_project_path}/{GlobalConfig.report_cfg.get('ALLURE_REPORT_PATH', default='report')}/allure".replace(
            "//", "/"
        )
        self.client_xml_report_path = lambda \
                x: f"/home/{x}/{self.server_project_path}/{GlobalConfig.report_cfg.get('XML_REPORT_PATH', default='report')}/xml".replace(
            "//", "/"
        )
        self.client_list = list(self.default.get(Args.clients.value).keys())
        _pty = "t"
        if len(self.client_list) >=2:
            _pty = "T"
        self.ssh = f"sshpass -p '%s' ssh -{_pty}"
        self.scp = "sshpass -p '%s' scp -r"
        self.rsync = "sshpass -p '%s' rsync -av -e ssh"
        self.empty = "> /dev/null 2>&1"
        self.strf_time = strftime("%m%d%p%I%M%S")

    def send_code_to_client(self, user, _ip, password):
        """
         发送代码到测试机
        :param user: 用户名
        :param _ip: 测试机IP
        :param password: 测试机密码
        :return:
        """
        logger.info(f"发送代码到测试机 - < {user}@{_ip} >")
        system(
            f"{self.ssh % password} {user}@{_ip} "
            f""""echo '{password}' | sudo -S rm -rf ~/{self.server_project_path}" {self.empty}"""
        )
        system(
            f'{self.ssh % password} {user}@{_ip} "mkdir -p ~/{self.server_project_path}" {self.empty}'
        )
        # 过滤目录
        app_name: str = self.default.get(Args.app_name.value)
        exclude = ""
        for i in [
            "report",
            "__pycache__",
            ".pytest_cache",
            ".vscode",
            ".idea",
            ".git",
            "docs",
            "README.md",
            "README.zh_CN.md",
            "RELEASE.md",
        ]:
            exclude += f"--exclude='{i}' "
        if app_name:
            for i in listdir(GlobalConfig.APPS_PATH):
                if i == "__init__.py":
                    continue
                if app_name.replace("-", "_") != i:
                    exclude += f"--exclude='{i}' "
        system(
            f"{self.rsync % (password,)} {exclude} {GlobalConfig.ROOT_DIR}/* "
            f"{user}@{_ip}:~/{self.server_project_path}/ {self.empty}"
        )
        logger.info(f"代码发送成功 - < {user}@{_ip} >")

    def build_client_env(self, user, _ip, password):
        """
         测试机环境安装
        :param user: 用户名
        :param _ip: 测试机IP
        :param password: 测试机密码
        :return:
        """
        logger.info(f"安装环境 - < {user}@{_ip} >")
        system(
            f"{self.ssh % password} {user}@{_ip} "
            f'''"rm -rf ~/.local/share/virtualenvs/{self.server_project_path.split('/')[-1]}*"'''
        )
        system(
            f"{self.ssh % password} {user}@{_ip} "
            f'"cd ~/{self.server_project_path}/ && bash env.sh"'
        )
        logger.info(f"环境安装完成 - < {user}@{_ip} >")

    def send_code_and_env(self, user, _ip, password):
        """
         发送代码到测试机并且安装环境
        :param user: 用户名
        :param _ip: 测试机IP
        :param password: 测试机密码
        :return:
        """
        self.send_code_to_client(user, _ip, password)
        self.build_client_env(user, _ip, password)

    def install_deb(self, user, _ip, password):
        """
         安装 deb 包
        :param user:
        :param _ip:
        :param password:
        :return:
        """
        logger.info(f"安装deb包 - < {user}@{_ip} >")
        system(
            f"{self.scp % password} {self.default.get(Args.deb_path.value)}/*.deb {user}@{_ip}:{self.default.get(Args.deb_path.value)}/"
        )
        system(
            f'''{self.ssh % password} {user}@{_ip} "cd {self.default.get(Args.deb_path.value)}/ && echo {password} | sudo -S dpkg -i *.deb"'''
        )
        logger.info(f"deb包安装完成 - < {user}@{_ip} >")

    def mul_do(self, func_obj, client_list):
        """
         异步发送代码
        :param func_obj: 函数对象
        :param client_list: 测试机列表
        :return:
        """
        if len(client_list) >= 2:
            executor = ThreadPoolExecutor()
            _ps = []
            for client in client_list[:-1]:
                user, _ip, password = self.default.get(Args.clients.value).get(client)
                _p1 = executor.submit(func_obj, user, _ip, password)
                _ps.append(_p1)
            user, _ip, password = self.default.get(Args.clients.value).get(client_list[-1])
            func_obj(user, _ip, password)
            wait(_ps, return_when=ALL_COMPLETED)
        else:
            user, _ip, password = self.default.get(Args.clients.value).get(client_list[0])
            func_obj(user, _ip, password)

    def get_client_test_status(self, user, _ip, password):
        """
         获取测试机是否有用例执行
        :param user: 用户名
        :param _ip: 测试机IP
        :param password: 测试机密码
        :return:
        """
        status_test = popen(
            f'{self.ssh % password} {user}@{_ip} "ps -aux | grep pytest | grep -v grep"'
        ).read()
        return bool(status_test)

    @staticmethod
    def make_dir(dirs):
        """make_dir"""
        if not exists(dirs):
            makedirs(dirs)

    def run_pytest_cmd(self, user, _ip, password):
        """
         创建 Pytest 命令行参数
        :param user: 用户名
        :param _ip: 测试机IP
        :param password: 测试机密码
        :return:
        """
        # pylint: disable=too-many-branches
        cmd = [
            self.ssh % password,
            f"{user}@{_ip}",
            '"',
            "cd",
        ]

        l_args = list(self.local_kwargs.items())
        real_app_name = ""
        _tmp_args = []
        for i in l_args:
            if i[1] is None:
                continue
            i = list(i)
            i[0] = f"--{i[0]}"
            i[1] = f"'{i[1]}'"
            if i[0] == "--app_name":
                real_app_name = f"apps/{self.default.get(Args.app_name.value)}"
                continue

            _tmp_args.extend(i)
        cmd.extend([
            f"~/{self.server_project_path}/{real_app_name}",
            "&&",
            "pipenv",
            "run",
        ])
        from src.rtk.local_runner import LocalRunner
        lr = LocalRunner(debug=True)
        lr_args = {k:v for k, v in lr.export_default.items() if v}
        rr_args = {k:v for k, v in self.local_kwargs.items() if v}
        lr_args.update(rr_args)
        pytest_cmd = lr.create_pytest_cmd(real_app_name.replace("apps/", ""), default=lr_args, proj_path=f"/home/{user}/{self.server_project_path}")

        cmd.extend(pytest_cmd)
        cmd.append('"')

        cmd_str = " ".join(cmd)
        logger.info(f"\n{cmd_str}\n")
        if self.default.get(Args.debug.value):
            logger.info("DEBUG 模式不执行用例!")
        else:
            system(cmd_str)

    def pytest_co_cmd(self):
        """
         创建收集用例的命令行参数
        :return:
        """
        app_dir = (
            self.default.get(Args.app_name.value)
            if self.default.get(Args.app_name.value)
            else ""
        )
        cmd = [
            "pytest",
            f"{GlobalConfig.APPS_PATH}/{app_dir}",
        ]
        if self.default.get(Args.keywords.value):
            cmd.extend(["-k", f"'{self.default.get(Args.keywords.value)}'"])
        if self.default.get(Args.tags.value):
            cmd.extend(["-m", f"'{self.default.get(Args.tags.value)}'"])
        if self.default.get(Args.noskip.value):
            cmd.extend(["--noskip", self.default.get(Args.noskip.value)])
        if self.default.get(Args.ifixed.value):
            cmd.extend(["--ifixed", self.default.get(Args.ifixed.value)])
        if self.default.get(Args.send_pms.value):
            cmd.extend(["--send_pms", self.default.get(Args.send_pms.value)])
        if self.default.get(Args.task_id.value):
            cmd.extend(["--task_id", self.default.get(Args.task_id.value)])
        if self.default.get(Args.trigger.value):
            cmd.extend(["--trigger", "auto"])
        cmd.append("--co")
        collect_only_cmd = " ".join(cmd)
        logger.info(f"Collecting: \n{collect_only_cmd}")
        collect_only_log = CmdCtl.run_cmd(collect_only_cmd)
        re_expr = compile(r"<Module (.*?\.py)>")
        cases = re.findall(re_expr, collect_only_log)
        if not cases:
            logger.error("未收集到用例")
            sys.exit(0)
        return cases

    def pre_env(self):
        """
         前置环境处理
        :return:
        """
        # rm hosts
        system(f"rm -rf ~/.ssh/known_hosts {self.empty}")
        # rm server report
        if self.clean_server_report_dir:
            system(f"rm -rf {GlobalConfig.REPORT_PATH}/* {self.empty}")
        # rm client report
        if not self.default.get(Args.send_code.value) and self.clean_client_report_dir:
            for client in self.default.get(Args.clients.value):
                user, _ip, password = self.default.get(Args.clients.value).get(client)
                system(
                    f"""{self.ssh % password} {user}@{_ip} "rm -rf {self.client_report_path(user)}/*" {self.empty}"""
                )
        # delete ssh ask
        sudo = f"echo '{GlobalConfig.PASSWORD}' | sudo -S"
        if "StrictHostKeyChecking no" not in popen("cat /etc/ssh/ssh_config").read():
            system(
                f"""{sudo} sed -i "s/#   StrictHostKeyChecking ask/ StrictHostKeyChecking no/g" /etc/ssh/ssh_config {self.empty}"""
            )
        # install sshpass
        if "(C)" not in popen("sshpass -V").read():
            system(f"{sudo} apt update {self.empty}")
            system(f"{sudo} apt install sshpass {self.empty}")

    def scp_report(self, user, _ip, password):
        """
         远程复制报告
        :param user:
        :param _ip:
        :param password:
        :return:
        """
        server_allure_path = f"{GlobalConfig.REPORT_PATH}/allure/{self.strf_time}_ip{_ip}_{self.default.get(Args.app_name.value)}"
        self.make_dir(server_allure_path)
        system(
            f"{self.scp % password} {user}@{_ip}:{self.client_allure_report_path(user)}/* {server_allure_path}/ {self.empty}"
        )
        generate_allure_html = f"{server_allure_path}/html"

        AllureCustom.gen(server_allure_path, generate_allure_html)

    def get_report(self, client_list):
        """
         回传测试报告
        :param client_list: 客户端列表
        :return:
        """
        # mul get report
        if len(self.default.get(Args.clients.value)) >= 2:
            _ps = []
            executor = ThreadPoolExecutor()
            for client in client_list[:-1]:
                user, _ip, password = self.default.get(Args.clients.value).get(client)
                _p4 = executor.submit(self.scp_report, user, _ip, password)
                _ps.append(_p4)
                sleep(2)
            user, _ip, password = self.default.get(Args.clients.value).get(client_list[-1])
            self.scp_report(user, _ip, password)
            wait(_ps, return_when=ALL_COMPLETED)
        else:
            user, _ip, password = self.default.get(Args.clients.value).get(client_list[0])
            self.scp_report(user, _ip, password)

    def parallel_run(self, client_list):
        """
         并行跑
        :param client_list:
        :return:
        """
        _ps = []
        executor = ThreadPoolExecutor()
        for client in client_list[:-1]:
            user, _ip, password = self.default.get(Args.clients.value).get(client)
            _p3 = executor.submit(self.run_pytest_cmd,  user, _ip, password)
            _ps.append(_p3)
            sleep(1)
        user, _ip, password = self.default.get(Args.clients.value).get(client_list[-1])
        self.run_pytest_cmd(
            user,
            _ip,
            password,
        )
        wait(_ps, return_when=ALL_COMPLETED)
        sleep(5)

    def nginx_run(self, client_list):
        """
         分布式执行
        :param client_list: 客户端列表
        :return:
        """
        # pylint: disable=too-many-nested-blocks
        case_files = self.pytest_co_cmd()
        logger.info(f"Collected {len(case_files)} case.")
        # sort case
        case_files.sort(key=lambda x: int(re.findall(r"(\d+)", x)[0]))
        _ps = []
        executor = ThreadPoolExecutor()
        for case in case_files:
            counter = {}
            try:
                # pylint: disable=unsubscriptable-object
                for client in cycle(client_list)[::-1]:
                    user, _ip, password = self.default.get(Args.clients.value).get(client)
                    if not self.get_client_test_status(user, _ip, password):
                        _p2 = executor.submit(
                            self.run_pytest_cmd,
                            user,
                            _ip,
                            password,
                            f"{splitext(case)[0]} and ({self.default.get(Args.keywords.value)})",
                            self.default.get(Args.tags.value),
                        )
                        _ps.append(_p2)
                        # relax and wait for pytest start
                        for _ in range(self.scan):
                            if self.get_client_test_status(user, _ip, password):
                                counter[client] = 0
                                break
                            # else:
                            sleep(1)
                        else:
                            client_list.remove(client)
                    else:
                        # relax
                        counter[client] = counter.get(client, 0) + 1
                        if counter.get(client) >= self.scan:
                            client_list.remove(client)
                        sleep(1)
            except ValueError:
                break
        # Wait for all child processes to end
        wait(_ps, return_when=ALL_COMPLETED)

    def remote_run(self):
        """
         远程执行主函数
        :return:
        """
        client_list = list(self.default.get(Args.clients.value).keys())
        self.pre_env()
        logger.info(
            "\n测试机列表:\n" + "\n".join([str(i) for i in self.default.get(Args.clients.value).items()])
        )
        if self.default.get(Args.build_env.value):
            self.mul_do(self.send_code_and_env, client_list)
        else:
            if self.default.get(Args.send_code.value):
                self.mul_do(self.send_code_to_client, client_list)
        if self.default.get(Args.deb_path.value):
            self.mul_do(self.install_deb, client_list)
        if self.default.get(Args.parallel.value):
            self.parallel_run(client_list)
        else:
            self.nginx_run(client_list)
        # collect and integrate result data after all tests.
        self.get_report(client_list)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import time
import os
import pwd
import errno
import re
import subprocess
import tempfile
import random
import glob
from .config import config


def scratchFile(label):  # pragma: no cover
    """Uses tempfile.NamedTemporaryFile() to create a unique tempfile in
     config.scratchDir, with a filename like:
     dogtail-headless-<label>.<random junk>"""
    prefix = "dogtail-headless-"
    return tempfile.NamedTemporaryFile(prefix="%s%s." % (prefix, label),
                                       dir=config.scratchDir, mode='w')


def testBinary(path):  # pragma: no cover
    if (path.startswith(os.path.sep) or
            path.startswith(os.path.join('.', '')) or
            path.startswith(os.path.join('..', ''))):
        if not os.path.exists(path):
            raise IOError(errno.ENOENT, "No such file", path)
        if not os.access(path, os.X_OK):
            raise IOError(errno.ENOEXEC, "Permission denied", path)
    return True


def get_username():  # pragma: no cover
    return pwd.getpwuid(os.getuid())[0]


class Subprocess(object):  # pragma: no cover

    def __init__(self, cmdList, environ=None):
        testBinary(cmdList[0])
        self.cmdList = cmdList
        self.environ = environ
        self._exitCode = None

    def start(self):
        if self.environ is None:
            self.environ = os.environ
        self.popen = subprocess.Popen(
            self.cmdList, env=self.environ)
        return self.popen.pid

    def wait(self):
        return self.popen.wait()

    def stop(self):
        self.popen.terminate()

    @property
    def exitCode(self):
        if self._exitCode is None:
            self._exitCode = self.wait()
        return self._exitCode


class XServer(Subprocess):  # pragma: no cover

    def __init__(self, server="/usr/bin/Xorg",
                 xinitrc="/etc/X11/xinit/Xclients",
                 resolution="1024x768x16"):
        """resolution is only used with Xvfb."""
        testBinary(server)
        self.server = server
        self._exitCode = None
        self.xinit = "/usr/bin/xinit"
        self.display = None
        self.xinitrc = xinitrc
        self.resolution = resolution

    @staticmethod
    def findFreeDisplay():
        tmp = os.listdir('/tmp')
        pattern = re.compile('\.X([0-9]+)-lock')
        usedDisplays = []
        for file in tmp:
            match = re.match(pattern, file)
            if match:
                usedDisplays.append(int(match.groups()[0]))
        if not usedDisplays:
            return ':0'
        usedDisplays.sort()
        return ':' + str(usedDisplays[-1] + 1)

    @property
    def cmdList(self):
        self.display = self.findFreeDisplay()
        cmd = []
        if self.xinit:
            cmd.append(self.xinit)
            if self.xinitrc:
                cmd.append(self.xinitrc)
            cmd.append('--')
        cmd.append(self.server)
        cmd.append(self.display)
        cmd.extend(['-ac', '-noreset'])
        if self.server.endswith('Xvfb'):
            cmd.extend(['-screen', '0', self.resolution])
            cmd.append('-shmem')
        return cmd

    def start(self):
        print(' '.join(self.cmdList))
        self.popen = subprocess.Popen(self.cmdList)
        return self.popen.pid


class Script(Subprocess):  # pragma: no cover
    pass


class Session(object):  # pragma: no cover

    cookieName = "DOGTAIL_SESSION_COOKIE"

    def __init__(self, sessionBinary, scriptCmdList=[], scriptDelay=20, logout=True):
        testBinary(sessionBinary)
        self.sessionBinary = sessionBinary
        self.script = Script(scriptCmdList)
        self.scriptDelay = scriptDelay
        self.logout = logout
        self.xserver = XServer()
        self._cookie = None
        self._environment = None

    def start(self):
        self.xinitrcFileObj = scratchFile('xinitrc')
        self.xserver.xinitrc = self.xinitrcFileObj.name
        self._buildXInitRC(self.xinitrcFileObj)
        xServerPid = self.xserver.start()
        time.sleep(self.scriptDelay)
        self.script.environ = self.environment
        scriptPid = self.script.start()
        return (xServerPid, scriptPid)

    @property
    def environment(self):
        def isSessionProcess(fileName):
            try:
                if os.path.realpath(path + 'exe') != ('/usr/bin/plasma-desktop'
                                                      if self.sessionBinary.split('/')[-1] == 'startkde'
                                                      else self.sessionBinary):
                    return False
            except OSError:
                return False
            pid = fileName.split('/')[2]
            if pid == 'self' or pid == str(os.getpid()):
                return False
            return True

        def getEnvDict(fileName):
            try:
                with open(fileName, 'r') as f:
                    envString = f.read()
            except IOError:
                return {}
            envItems = envString.split('\x00')
            envDict = {}
            for item in envItems:
                if '=' not in item:
                    continue
                k, v = item.split('=', 1)
                envDict[k] = v
            return envDict

        def isSessionEnv(envDict):
            if not envDict:
                return False
            if envDict.get(self.cookieName, 'notacookie') == self.cookie:
                return True
            return False

        for path in glob.glob('/proc/*/'):
            if not isSessionProcess(path):
                continue
            envFile = path + 'environ'
            envDict = getEnvDict(envFile)
            if isSessionEnv(envDict):
                # print path
                # print envDict
                self._environment = envDict
        if not self._environment:
            raise RuntimeError("Can't find our environment!")
        return self._environment

    def wait(self):
        self.script.wait()
        return self.xserver.wait()

    def stop(self):
        try:
            self.script.stop()
        except OSError:
            pass
        self.xserver.stop()

    def attemptLogout(self):
        logoutScript = Script('dogtail-logout',
                              environ=self.environment)
        logoutScript.start()
        logoutScript.wait()

    @property
    def cookie(self):
        if not self._cookie:
            self._cookie = "%X" % random.getrandbits(16)
        return self._cookie

    def _buildXInitRC(self, fileObj):
        lines = [
            "export %s=%s" % (self.cookieName, self.cookie),
            "gsettings set org.gnome.desktop.interface toolkit-accessibility true",
            ". /etc/X11/xinit/xinitrc-common",
            "export %s" % self.cookieName,
            "exec -l $SHELL -c \"$CK_XINIT_SESSION $SSH_AGENT %s\"" %
            (self.sessionBinary),
            ""]

        fileObj.write('\n'.join(lines).strip())
        fileObj.flush()



#!/usr/bin/env python3
# _*_ coding:utf-8 _*_

# SPDX-FileCopyrightText: 2023 UnionTech Software Technology Co., Ltd.

# SPDX-License-Identifier: GPL-2.0-only
import os
from urllib.parse import urlencode

from setting import conf
from setting.globalconfig import FixedCsvTitle
from src.pms._base import _Base
from src.rtk._base import transform_app_name


class Csv2Pms(_Base):
    __author__ = "mikigo<huangmingqiang@uniontech.com>"

    config_error_log = "请检查您传递的 '命令行参数' 或 setting/globalconfig.ini 里的配置项"

    def __init__(
            self,
            app_name: str = None,
            user: str = None,
            password: str = None,
            csv_name=None
    ):
        super().__init__(user=user, password=password)
        self.username = user
        self.password = user
        self.base_url = "https://pms.uniontech.com"
        self.walk_dir = (
            f"{conf.APPS_PATH}/{transform_app_name(app_name)}"
            if app_name
            else conf.APPS_PATH
        )
        self.csv_name = csv_name or conf.CSV_NAME_TO_PMS

        if not self.csv_name:
            raise EnvironmentError(self.config_error_log)

    def get_csv_info(self):
        csv_path = None
        for root, _, files in os.walk(self.walk_dir):
            for file in files:
                if f"{self.csv_name}.csv" == file:
                    csv_path = f"{root}/{file}"
        if csv_path is None:
            raise FileNotFoundError(self.config_error_log)

        with open(csv_path, "r", encoding="utf-8") as f:
            txt_list = f.readlines()

        csv_heads = txt_list[0].strip().split(",")
        csv_head_index_map = {}
        for index, title in enumerate(csv_heads):
            for i in FixedCsvTitle:
                if i.value == title.strip():
                    csv_head_index_map[i.name] = {
                        "head_name": i.value,
                        "head_index": index,
                    }

        taglines = [txt.strip().split(",") for txt in txt_list[1:]]
        return csv_head_index_map, taglines

    def post_to_pms(self):
        csv_head_index_map, taglines = self.get_csv_info()

        def csv_map(x):
            return csv_head_index_map.get(x).get('head_index')

        for i in taglines:
            case_id = i[csv_map(FixedCsvTitle.case_id.name)]
            edit_url = f"{self.base_url}/testcase-edit-{case_id}.html"
            data = {
                'isAutomation': '是',
            }
            if i[csv_map(FixedCsvTitle.device_type.name)] in ("PPL", "COL"):
                data["deviceType"] = i[csv_map(FixedCsvTitle.device_type.name)]
            if i[csv_map(FixedCsvTitle.case_from.name)] == "是":
                data["caseSource"] = "是"
            if i[csv_map(FixedCsvTitle.online_obj.name)] == "CICD":
                data["lineCD"] = "是"
            bytes_data = urlencode(data).encode("utf-8")
            res = self.rx.session.open(
                fullurl=edit_url,
                data=bytes_data,
                timeout=10
            )
            print(f"{case_id}-{data}—{res.status}")



from public.dde_desktop_public_widget.dde_desktop_public_widget import DdeDesktopPublicWidget
from public.dde_dock_public_widget.dde_dock_public_widget import DdeDockPublicWidget
from public.dde_launcher_public_widget.dde_launcher_public_widget import DdeLauncherPublicWidget
from public.open_mode_public_widget.open_mode_public_widget import OpenModePublicWidget
from public.other_public_widget.deepin_compress_public_widget import DeepinCompressPublicWidget
from public.other_public_widget.deepin_manual_public_widget import DeepinManualPublicWidget
from public.right_menu_public_widget.right_menu_public_widget import RightMenuPublicWidget



#!/usr/bin/env python3
# _*_ coding:utf-8 _*_

# SPDX-FileCopyrightText: 2023 UnionTech Software Technology Co., Ltd.

# SPDX-License-Identifier: GPL-2.0-only
# pylint: disable=C0301,R0912,C0413,R0914,W0212,R1702,R0915
# pylint: disable=C0114,W0621,C0411,C0412,R1706,E0401
import sys
from os import environ

environ["DISPLAY"] = ":0"
from setting.globalconfig import SystemPath

for i in SystemPath:
    if i.value not in sys.path:
        sys.path.append(i.value)

from os import popen
from os import system
from os import remove
from os import makedirs
from os import walk
from os.path import exists
from os.path import splitext
from enum import Enum
from time import sleep
from collections import deque
from datetime import datetime
from json import dumps
from re import findall
from shutil import copyfile
from multiprocessing import Process
from concurrent.futures import wait
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import ALL_COMPLETED

import letmego
import allure
import pytest
from _pytest.mark import Mark
from _pytest.terminal import TerminalReporter
from funnylog.conf import setting as log_setting

from setting.globalconfig import GlobalConfig

letmego.conf.setting.PASSWORD = GlobalConfig.PASSWORD
letmego.conf.setting.RUNNING_MAN_FILE = f"{GlobalConfig.REPORT_PATH}/_running_man.log"
letmego.conf.setting.DEBUG = GlobalConfig.LETMEGO_DEBUG

log_setting.LOG_FILE_PATH = GlobalConfig.REPORT_PATH
log_setting.CLASS_NAME_STARTSWITH = GlobalConfig.CLASS_NAME_STARTSWITH
log_setting.CLASS_NAME_ENDSWITH = GlobalConfig.CLASS_NAME_ENDSWITH
log_setting.CLASS_NAME_CONTAIN = GlobalConfig.CLASS_NAME_CONTAIN

from setting import skipif
from setting.globalconfig import ConfStr
from setting.globalconfig import FixedCsvTitle
from src import logger
from src.plugins.allure_report_extend import AllureReportExtend
from src.plugins import emoji_hooks
from src.cmdctl import CmdCtl
from src.pms._base import write_case_result
from src.pms._base import runs_id_cmd_log
from src.pms.task import Task
from src.pms.suite import Suite
from src.pms.send2pms import Send2Pms
from src.recording_screen import recording_screen

FLAG_FEEL = "=" * 10
LN = "\n"


class LabelType(Enum):
    """用例级别对应报告级别"""

    L1 = allure.severity_level.BLOCKER
    L2 = allure.severity_level.CRITICAL
    L3 = allure.severity_level.NORMAL
    L4 = allure.severity_level.MINOR


def add_mark(item, name: str = "", args: tuple = (), kwargs: dict = None):
    """add mark"""
    item.own_markers.append(Mark(name=name, args=args, kwargs=kwargs))


def write_json(session):
    """write json"""
    return bool(
        session.config.option.send_pms
        and (session.config.option.task_id or session.config.option.suite_id)
    )


def auto_send(session):
    """auto send"""
    return bool(session.config.option.send_pms and session.config.option.trigger)


def async_send(session):
    """async send"""
    return bool(
        session.config.option.send_pms == ConfStr.ASYNC.value
        and session.config.option.trigger == ConfStr.AUTO.value
    )


def finish_send(session):
    """finish send"""
    return bool(
        session.config.option.send_pms == ConfStr.FINISH.value
        and session.config.option.trigger == ConfStr.AUTO.value
    )


def pytest_addoption(parser):
    """pytest_cmdline_main"""
    parser.addoption(
        "--clean", action="store", default="no", help="是否清理环境&杀进程"
    )
    parser.addoption(
        "--log_level", action="store", default=GlobalConfig.LOG_LEVEL, help="终端日志输出级别"
    )
    parser.addoption(
        "--noskip", action="store", default="", help="skip-xxx标签不生效"
    )
    parser.addoption(
        "--ifixed", action="store", default="", help="fixed-xxx标签不生效"
    )
    parser.addoption(
        "--max_fail", action="store", default="", help="最大失败次数"
    )
    parser.addoption(
        "--record_failed_case", action="store", default="",
        help="失败录屏从第几次失败开始录制视频"
    )
    parser.addoption(
        "--send_pms", action="store", default="", help="用例数据回填"
    )
    parser.addoption(
        "--task_id", action="store", default="", help="测试单id"
    )
    parser.addoption(
        "--trigger", action="store", default="", help="数据回填的触发者"
    )
    parser.addoption(
        "--suite_id", action="store", default="", help="pms的测试套件ID"
    )
    parser.addoption(
        "--pms_user", action="store", default="", help="登录pms的账号"
    )
    parser.addoption(
        "--pms_password", action="store", default="", help="登录pms的密码"
    )
    parser.addoption(
        "--top", action="store", default="", help="过程中记录top命令中的值"
    )
    parser.addoption(
        "--duringfail", action="store_true", dest="duringfail", default=False,
        help="出现错误时立即显示",
    )
    parser.addoption(
        "--repeat", action="store", default=1, type=int, help="用例重复执行的次数"
    )
    parser.addoption(
        "--export_csv_file", action="store", default="", help="导出csv文件"
    )
    parser.addoption(
        "--line", action="store", default="", help="业务线(CI)"
    )
    parser.addoption(
        "--app_name", action="store", default="", help="执行的应用名称"
    )
    parser.addoption(
        "--autostart", action="store", default="", help="重启类场景开启letmego执行方案"
    )


def pytest_cmdline_main(config):
    """pytest_cmdline_main"""
    # 初始化log配置，以解决allure报告日志格式问题
    log_info = logger(config.option.log_level)
    config.option.log_level = config.option.log_level
    config.option.log_format = log_info.log_format
    config.option.log_date_format = log_info.date_format


def pytest_addhooks(pluginmanager):
    """pytest_addhooks"""
    pluginmanager.add_hookspecs(emoji_hooks)


@pytest.mark.trylast
def pytest_configure(config):
    """pytest_configure"""
    if hasattr(config, "workerinput"):
        return  # xdist worker
    # 获取终端报告器插件
    reporter = config.pluginmanager.getplugin("terminalreporter")
    if config.option.duringfail and reporter:
        custom_reporter = DuringfailingTerminalReporter(reporter)
        # 替换成我们自己的报告插件
        config.pluginmanager.unregister(custom_reporter)
        config.pluginmanager.register(custom_reporter)


def pytest_sessionstart(session):
    """pytest_sessionstart"""
    if (
            CmdCtl.run_cmd(
                "gsettings get com.deepin.dde.appearance gtk-theme",
                interrupt=False,
                out_debug_flag=False,
                command_log=False,
            ).strip("'")
            != GlobalConfig.SYS_THEME
    ):
        CmdCtl.run_cmd(
            f"gsettings set com.deepin.dde.appearance gtk-theme {GlobalConfig.SYS_THEME}",
            interrupt=False,
            out_debug_flag=False,
            command_log=False,
        )
    _display = (
        GlobalConfig.DisplayServer.wayland
        if GlobalConfig.IS_WAYLAND
        else GlobalConfig.DisplayServer.x11
    )
    logger.info(f"当前系统显示协议为 {_display.title()}")
    popen("gsettings set com.deepin.dde.dock position bottom")
    session.config.option.start_time = datetime.now()

    user = session.config.option.pms_user
    password = session.config.option.pms_password
    task_id = session.config.option.task_id
    suite_id = session.config.option.suite_id
    if write_json(session):
        session.case_res_path = Send2Pms.case_res_path(task_id or suite_id)
        session.data_send_result_csv = Send2Pms.data_send_result_csv(
            task_id or suite_id
        )

    if user and password and async_send(session):
        session.all_thread_task = []
        session.t_executor = ThreadPoolExecutor()

    if not session.config.option.collectonly and session.config.option.top:

        def record_top():
            top_log_path = f"{GlobalConfig.REPORT_PATH}/logs"
            if not exists(top_log_path):
                makedirs(top_log_path)
            system(
                f"{GlobalConfig.top_cmd} | grep ^top -A {int(session.config.option.top) + 6} > "
                f"{top_log_path}/top_{GlobalConfig.TIME_STRING}.log"
            )

        session.p = Process(target=record_top, args=())
        session.p.start()


@pytest.hookimpl(trylast=True)
def pytest_generate_tests(metafunc):
    """pytest generate tests"""
    repeat = metafunc.config.option.repeat
    marks = metafunc.definition.get_closest_marker("repeat")
    if marks is not None:
        repeat = int(marks.args[0])
    if repeat > 1:
        metafunc.fixturenames.append("__pytest_repeat_step_number")

        def ids(i, number=repeat):
            return f"{i + 1}-{number}"

        metafunc.parametrize(
            "__pytest_repeat_step_number",
            range(repeat),
            indirect=True,
            ids=ids,
        )


def pytest_collection_modifyitems(session):
    """pytest collection modifyitems"""

    walk_dir = (
        f"{GlobalConfig.APPS_PATH}/{session.config.option.app_name}"
        if session.config.option.app_name and exists(f"{GlobalConfig.APPS_PATH}/{session.config.option.app_name}")
        else GlobalConfig.APPS_PATH
    )
    csv_path_dict, no_youqu_mark = walk_apps(walk_dir)

    if not csv_path_dict:
        return

    user = session.config.option.pms_user
    password = session.config.option.pms_password
    suite_id = session.config.option.suite_id
    task_id = session.config.option.task_id
    containers = {}
    suite_runs_ids = suit_id_deque = task_runs_ids = task_id_deque = None
    skip_index = fixed_index = removed_index = pms_id_index = None

    if suite_id and task_id:
        raise ValueError("suite_id 和 task_id 不能同时存在~")
    if suite_id or task_id:
        if not (user and password):
            raise ValueError("pms_user 或 pms_password 未传入")
    if suite_id:
        suite_runs_ids, suit_id_deque = get_runs_id_deque(
            user, password, Suite, "suite", suite_id
        )
        print(
            f"{LN}测试套件: https://pms.uniontech.com/zentao/testsuite-view-{suite_id}.html"
            f"{LN}关联的用例:{LN}{f'{LN}'.join([runs_id_cmd_log(i) for i in suite_runs_ids])}"
        )
    elif task_id:
        task_runs_ids, task_id_deque = get_runs_id_deque(
            user, password, Task, "task", task_id
        )
        print(
            f"{LN}测试单: https://pms.uniontech.com/testtask-cases-{task_id}.html"
            f"{LN}关联的用例:{LN}{f'{LN}'.join([runs_id_cmd_log(i) for i in task_runs_ids])}"
        )

    for item in session.items[::-1]:
        item.name = item.name.encode("utf-8").decode("unicode_escape")
        item._nodeid = item.nodeid.encode("utf-8").decode("unicode_escape")

        if no_youqu_mark:
            continue_flag = False
            for app_abspath in no_youqu_mark:
                if app_abspath in item.fspath.strpath:
                    continue_flag = True
                    break
            if continue_flag:
                continue

        try:
            csv_name, _id = findall(r"test_(.*?)_(\d+)", item.name)[0]
            _case_id = findall(r"test_.*?_(\d+)", item.fspath.purebasename)[0]
            if _id != _case_id:
                raise ValueError
        except IndexError:
            skip_text = f"\n用例函数名称缺少用例id:[{item.nodeid}]"
            logger.error(skip_text)
            add_mark(item, ConfStr.SKIP.value, (skip_text,), {})
        except ValueError:
            skip_text = f"\n用例py文件的id与用例函数的id不一致:[{item.nodeid}]"
            logger.error(skip_text)
            add_mark(item, ConfStr.SKIP.value, (skip_text,), {})
        else:
            csv_path = csv_path_dict.get(csv_name)
            if not csv_path:
                if "asan" not in csv_name:
                    logger.error(f"{csv_name}.csv 文件不存在!")
                continue

            if not containers.get(csv_path):
                with open(csv_path, "r", encoding="utf-8") as _f:
                    txt_list = _f.readlines()
                if not txt_list:
                    continue
                # 通过csv的表头找到对应的索引（排除ID列的索引）
                for index, title in enumerate(txt_list[0].strip().split(",")):
                    if title.strip() == FixedCsvTitle.skip_reason.value:
                        skip_index = index - 1
                    elif title.strip() == FixedCsvTitle.fixed.value:
                        fixed_index = index - 1
                    elif title.strip() == FixedCsvTitle.removed.value:
                        removed_index = index - 1
                    elif title.strip() == FixedCsvTitle.pms_case_id.value:
                        pms_id_index = index - 1

                taglines = [txt.strip().split(",") for txt in txt_list[1:]]
                id_tags_dict = {f"{int(i[0]):0>3}": i[1:] for i in taglines if i[0]}
                # 每个csv文件单独管理一套index
                containers[csv_path] = id_tags_dict
                containers[csv_path][ConfStr.SKIP_INDEX.value] = skip_index
                containers[csv_path][ConfStr.FIXED_INDEX.value] = fixed_index
                containers[csv_path][ConfStr.REMOVED_INDEX.value] = removed_index
                containers[csv_path][ConfStr.PMS_ID_INDEX.value] = pms_id_index
                # 将index重置
                skip_index = fixed_index = removed_index = pms_id_index = None
            # 如果是想通过测试单跑或者测试套件跑用例，但是csv文件里面又没有保存“PMS用例ID”列，直接不跑
            if (task_id or suite_id) and containers[csv_path][
                ConfStr.PMS_ID_INDEX.value
            ] is None:
                session.items.remove(item)
                continue
            tags = containers.get(csv_path).get(_id)
            if tags:
                try:
                    if containers[csv_path][
                        ConfStr.REMOVED_INDEX.value
                    ] is not None and tags[
                        containers[csv_path][ConfStr.REMOVED_INDEX.value]
                    ].strip(
                        '"'
                    ).startswith(
                        f"{ConfStr.REMOVED.value}-"
                    ):
                        session.items.remove(item)
                        continue
                except IndexError as exc:
                    logger.error(
                        f"\ncsv_path:\t{csv_path}\ntags:\t{tags}\n"
                        f"error_tag_index:\t{containers[csv_path][ConfStr.REMOVED_INDEX.value]}"
                    )
                    raise IndexError from exc
                for index, tag in enumerate(tags):
                    if tag:
                        tag = tag.strip('"')
                        # 先处理“跳过原因”列
                        if index == containers[csv_path][ConfStr.SKIP_INDEX.value]:
                            # 标签是以 “skip-” 开头, noskip 用于解除所有的skip
                            if not session.config.option.noskip and tag.startswith(
                                    f"{ConfStr.SKIP.value}-"
                            ):
                                # 标签以 “fixed-” 开头, ifixed表示ignore fixed, 用于忽略所有的fixed
                                # 1. 不给ifixed参数时，只要标记了fixed的用例，即使标记了skip-，也会执行；
                                # 2. 给ifixed 参数时(--ifixed yes)，fixed不生效，仅通过skip跳过用例；
                                try:
                                    if (
                                            not session.config.option.ifixed
                                            and containers[csv_path][
                                        ConfStr.FIXED_INDEX.value
                                    ]
                                            is not None
                                            and tags[
                                        containers[csv_path][
                                            ConfStr.FIXED_INDEX.value
                                        ]
                                    ]
                                            .strip('"')
                                            .startswith(f"{ConfStr.FIXED.value}-")
                                    ):
                                        continue
                                except IndexError:
                                    # 如果访问越界，说明这行没有fixed标签或者标签写错位置了，所以正常跳过
                                    pass
                                add_mark(item, ConfStr.SKIP.value, (tag,), {})
                            elif f"{ConfStr.SKIPIF.value}_" in tag:
                                skip_method, param = tag.split("-", maxsplit=1)
                                if hasattr(skipif, skip_method):
                                    skip_result = getattr(skipif, skip_method)(param)
                                    add_mark(
                                        item,
                                        ConfStr.SKIPIF.value,
                                        (skip_result,),
                                        {"reason": tag},
                                    )
                                else:
                                    logger.error(f"未找到判断是否跳过的自定义方法 <{skip_method}>")
                                    add_mark(
                                        item,
                                        ConfStr.SKIP.value,
                                        (f"未找到判断是否跳过的自定义方法 <{skip_method}>",),
                                        {},
                                    )
                        else:  # 非跳过列
                            # 处理 pms id
                            if (
                                    containers[csv_path][ConfStr.PMS_ID_INDEX.value]
                                    == index
                            ):
                                if suite_runs_ids:
                                    if tag not in suit_id_deque:
                                        session.items.remove(item)
                                        continue
                                    add_run_case_id(session, item, tag, suite_runs_ids)
                                elif task_runs_ids:
                                    if tag not in task_id_deque:
                                        session.items.remove(item)
                                        continue
                                    add_run_case_id(session, item, tag, task_runs_ids)

                            # 处理其他自定义标签
                            try:
                                mark_title = txt_list[0].strip().split(",")[index + 1]
                            except IndexError:
                                # 如果写了标签，但是没有对应的表头
                                mark_title = ""
                            add_mark(item, tag, (mark_title,), {})
                    else:  # tag为空
                        # 处理 pmd id 为空的情况
                        if (task_id or suite_id) and containers[csv_path][
                            ConfStr.PMS_ID_INDEX.value
                        ] == index:
                            session.items.remove(item)
                            continue
            else:
                if session.config.option.allure_report_dir:
                    # 批量执行时，不执行没有ID的用例。
                    logger.error(f"<{item.name}> csv文件中未标记,强制跳过")
                    session.items.remove(item)

    if session.config.option.autostart:
        for item in session.items[::-1]:
            _reruns = None
            if hasattr(session.config.option, "reruns"):
                _reruns = session.config.option.reruns
            if letmego.read_testcase_running_status(item, reruns=_reruns):
                session.items.remove(item)

    if (suite_id or task_id) and session.items:
        print("\n即将执行的用例:")
        for item in session.items:
            for mark in item.own_markers:
                if mark.args == (FixedCsvTitle.pms_case_id.value,):
                    print(f"case_id: {mark.name}, case_name: {item.name}")
                    break
        print()  # 处理日志换行


def pytest_collection_finish(session):
    """pytest collection finish"""
    session.item_count = len(session.items)
    print(f"用例收集数量:\t{session.item_count}")
    print(f"用例文件数量:\t{len(set([item.fspath for item in session.items]))}")
    if session.config.option.reruns and not session.config.option.collectonly:
        print(f"失败重跑次数:\t{session.config.option.reruns}")
    if session.config.option.max_fail and not session.config.option.collectonly:
        session.config.option.maxfail = int(
            float(session.config.option.max_fail) * session.item_count
        )
        print(f"最大失败次数:\t{session.config.option.maxfail}")
    session.sessiontimeout = 0
    if session.config.option.timeout and not session.config.option.collectonly:
        _min, sec = divmod(int(session.config.option.timeout), 60)
        hour, _min = divmod(_min, 60)
        print(
            f"用例超时时间:\t{session.config.option.timeout}s ({hour}{'小时' if hour else ''}{_min}{'分' if _min else ''}{sec}秒)"
        )
        # sessiontimeout
        _n = 0
        items_timeout = 0
        for item in session.items:
            for mark in item.own_markers:
                if mark.name == "timeout":
                    try:
                        item_timeout = mark.args[0]
                        _n += 1
                    except IndexError:
                        item_timeout = 0
                    items_timeout += item_timeout
                    break
        session.sessiontimeout = (
                                         (session.item_count - _n) * session.config.option.timeout
                                 ) + items_timeout
        _min, sec = divmod(int(session.sessiontimeout), 60)
        hour, _min = divmod(_min, 60)
        print(
            f"会话超时时间:\t{session.sessiontimeout}s ({hour}{'小时' if hour else ''}{_min}{'分' if _min else ''}{sec}秒)"
        )

    if session.config.option.collectonly and session.config.option.export_csv_file:
        execute = []
        execute.append("用例名称," + GlobalConfig.EXPORT_CSV_HEARD + LN)
        for item in session.items:
            node_id = item.nodeid.split("[")[0]
            header = GlobalConfig.EXPORT_CSV_HEARD.split(",")
            case_info = ["" for _ in header]
            case_info.insert(0, node_id)
            for mark in item.own_markers:
                try:
                    index = header.index(mark.args[0]) + 1
                except (ValueError, IndexError):
                    continue
                case_info[index] = mark.name
            execute.append(",".join(case_info) + LN)
        execute2 = list(set(execute))
        execute2.sort(key=execute.index)
        if not exists(GlobalConfig.REPORT_PATH):
            makedirs(GlobalConfig.REPORT_PATH)
        with open(
                f"{GlobalConfig.REPORT_PATH}/{session.config.option.export_csv_file}",
                "w+",
                encoding="utf-8",
        ) as _f:
            _f.writelines(execute2)


def pytest_runtest_setup(item):
    """pytest runtest setup"""
    if hasattr(item, "execution_count"):
        letmego.conf.setting.EXECUTION_COUNT = item.execution_count

    print()  # 处理首行日志换行的问题
    current_item_count = (
        f"[{item.session.items.index(item) + 1}/{item.session.item_count}]"
    )
    try:
        rerun_text = (
            f" | <重跑第{item.execution_count - 1}次>" if item.execution_count > 1 else ""
        )
    except AttributeError:
        rerun_text = ""
    logger.info(
        f"{LN}{FLAG_FEEL} {item.function.__name__} || "
        f"{str(item.function.__doc__).replace(LN, '').replace('    ', '')}{rerun_text} "
        f"{FLAG_FEEL} {current_item_count}"
    )
    try:
        if item.execution_count >= (int(item.config.option.record_failed_case) + 1):
            logger.info("开启录屏")
            item.record = {}
            item.record["object"] = recording_screen(
                f"{item.name}_{item.execution_count}"
            )  # 存放录屏对象
            item.record["image_path"] = next(item.record["object"])  # 录屏文件地址
            sleep(3)  # 等待3秒，优化录屏效果
    except AttributeError:
        pass

    if item.config.option.pms_user and item.config.option.pms_password:

        def send2pms(case_res_path, data_send_result_csv):
            Send2Pms(
                user=item.config.option.pms_user,
                password=item.config.option.pms_password,
            ).send2pms(case_res_path, data_send_result_csv)

        if async_send(item.session):
            task = item.session.t_executor.submit(
                send2pms, item.session.case_res_path, item.session.data_send_result_csv
            )
            item.session.all_thread_task.append(task)


# pylint: disable=unused-argument
def pytest_runtest_call(item):
    """pytest runtest call"""
    logger.info(f"{FLAG_FEEL} case body {FLAG_FEEL}")


def pytest_runtest_teardown(item):
    """pytest runtest teardown"""
    logger.info(f"{FLAG_FEEL} teardown {FLAG_FEEL}")
    sessiontimeout = item.session.sessiontimeout
    if sessiontimeout:
        duration = datetime.now() - item.session.config.option.start_time
        if duration.seconds > int(sessiontimeout):
            # 处理时间秒为 XX分XX秒
            _min, sec = divmod(duration.seconds, 60)
            # 处理时间分为 XX小时xx分xx秒
            hour, _min = divmod(_min, 60)
            raise item.session.Interrupted(f"会话超时（{hour}小时{_min}分{sec}秒）,用例强制终止!")


@pytest.hookimpl(hookwrapper=True, tryfirst=True)
def pytest_runtest_makereport(item, call):
    """pytest_runtest_makereport"""
    out = yield
    report = out.get_result()
    if report.when == "setup":
        for mark in item.own_markers:
            if mark.name == "parametrize":
                continue
            if mark.args[0] == FixedCsvTitle.case_level.value:
                try:
                    allure.dynamic.severity(LabelType[mark.name].value)
                except KeyError:
                    allure.dynamic.severity(LabelType.L3.value)
            elif mark.args[0] == FixedCsvTitle.pms_case_id.value:
                # if mark.name:
                testcase_url = (
                    f"https://pms.uniontech.com/testcase-view-{mark.name}.html"
                )
                allure.dynamic.testcase(testcase_url)
                logger.info(testcase_url)
            else:
                allure.dynamic.tag(mark.name)
    if report.when == "call":
        logger.info(f"运行结果: {str(report.outcome).upper()}")
        if write_json(item.session):
            # 只要是需要数据回填（无论是自动还是手动）,都需要写json结果.
            write_case_result(item, report)

        if item.config.option.autostart:
            letmego.write_testcase_running_status(item, report)
    try:
        if item.execution_count >= (int(item.config.option.record_failed_case) + 1):
            if report.when == "call":  # 存放录屏当次测试结果
                item.record["result"] = report.outcome
                try:
                    # 记录断言的模板图片
                    item.record["template"] = call.excinfo.value.args[0].args[1:]
                except (IndexError, KeyError, AttributeError):
                    # 记录ocr识别区域图片
                    try:
                        pic = call.excinfo.value.args[0][1]
                        if isinstance(pic, tuple):
                            item.record["ocr"] = call.excinfo.value.args[0][1]
                    except (IndexError, AttributeError, TypeError):
                        # 非ocr断言
                        pass
            elif report.when == "teardown":
                try:
                    sleep(3)
                    # 调用生成器保存视频
                    next(item.record["object"])
                except StopIteration:
                    # 录屏时测试结果为passed，则删除视频
                    if item.record.get("result") == ConfStr.PASSED.value:
                        try:
                            remove(item.record["image_path"])
                        except FileNotFoundError:
                            pass
                    else:
                        if exists(GlobalConfig.SCREEN_CACHE):
                            screen_png = f"{splitext(item.record['image_path'])[0]}.png"
                            copyfile(GlobalConfig.SCREEN_CACHE, screen_png)
                            allure.attach.file(
                                screen_png,
                                name="屏幕截图",
                                attachment_type=allure.attachment_type.PNG,
                            )
                            try:
                                for index, tem in enumerate(item.record["template"]):
                                    template = f"{splitext(item.record['image_path'])[0]}_template_{index}.png"
                                    CmdCtl.run_cmd(f"cp {tem} {template}")
                                    allure.attach.file(
                                        template,
                                        name="模板图片",
                                        attachment_type=allure.attachment_type.PNG,
                                    )
                            except KeyError:
                                # 非图像识别错误
                                pass
                            try:
                                template = (
                                    f"{splitext(item.record['image_path'])[0]}_ocr_.png"
                                )
                                CmdCtl.run_cmd(f"cp {item.record['ocr']} {template}")
                                allure.attach.file(
                                    template,
                                    name="OCR识别区域",
                                    attachment_type=allure.attachment_type.PNG,
                                )
                            except KeyError:
                                # ocr 识别区域
                                pass
                        allure.attach.file(
                            item.record["image_path"],
                            name="用例视频",
                            attachment_type=allure.attachment_type.MP4,
                        )
                    logger.info(
                        "结束录屏! "
                        f"{'重跑用例测试成功，删除视频录像' if item.record.get('result') == ConfStr.PASSED.value else ''}"
                    )
    except (AttributeError, KeyError):
        pass


def pytest_report_teststatus(report, config):
    """pytest report teststatus"""
    # 在 setup 和 teardown 阶段处理 error 和 skip
    if report.when in ("setup", "teardown"):
        if report.failed:
            short, verbose = config.hook.pytest_emoji_error(
                config=config, head_line=report.head_line
            )
            return "error", short, verbose
        if report.skipped:
            short, verbose = config.hook.pytest_emoji_skipped(
                config=config, head_line=report.head_line
            )
            return "skipped", short, verbose
    # 在用例执行阶段处理 passed skipped failed
    if report.when == "call":
        short = verbose = ""
        if report.passed:
            short, verbose = config.hook.pytest_emoji_passed(
                config=config, head_line=report.head_line
            )
        elif report.skipped:
            short, verbose = config.hook.pytest_emoji_skipped(
                config=config, head_line=report.head_line
            )
        elif report.failed:
            short, verbose = config.hook.pytest_emoji_failed(
                config=config, head_line=report.head_line
            )
        return report.outcome, short, verbose
    return None


def pytest_sessionfinish(session):
    """pytest session finish"""
    if session.config.option.allure_report_dir:
        AllureReportExtend.environment_info(session)
        terminalreporter = session.config.pluginmanager.get_plugin("terminalreporter")
        execute = {}
        for _, items in terminalreporter.stats.items():
            for item in items:
                default_result = {"result": "blocked", "longrepr": "None"}
                try:
                    if item.outcome == ConfStr.PASSED.value:
                        default_result["result"] = "pass"
                    elif item.outcome == ConfStr.SKIPPED.value:
                        default_result["result"] = "skip"
                    elif item.outcome == ConfStr.RERUN.value:
                        continue
                    else:
                        default_result["result"] = "fail"
                    item_name = item.nodeid.split("[")[0]
                    if not execute.get(item_name) or (
                            item.outcome != ConfStr.PASSED.value
                            and execute.get(item_name).get("result") == "pass"
                    ):
                        execute[item_name] = default_result
                except AttributeError:
                    pass
        if execute:
            with open(
                    f"{GlobalConfig.ROOT_DIR}/ci_result.json", "w", encoding="utf-8"
            ) as _f:
                _f.write(dumps(execute, indent=2, ensure_ascii=False))

    if session.config.option.pms_user and session.config.option.pms_password:

        def send2pms(case_res_path, data_send_result_csv):
            Send2Pms(
                user=session.config.option.pms_user,
                password=session.config.option.pms_password,
            ).send2pms(case_res_path, data_send_result_csv)

        if async_send(session):
            wait(session.all_thread_task, return_when=ALL_COMPLETED)
            send2pms(session.case_res_path, session.data_send_result_csv)
            session.t_executor.shutdown()

        if finish_send(session):
            send2pms(session.case_res_path, session.data_send_result_csv)

    if not session.config.option.collectonly and session.config.option.top:
        session.p.terminate()
        system(
            f"ps -aux | grep '{GlobalConfig.top_cmd}' | "
            "cut -c 9-15 | xargs kill -9 > /dev/null 2>&1"
        )
        session.p.close()

    if exists(GlobalConfig.TMPDIR):
        # 清理临时模板图片
        CmdCtl.run_cmd(
            f"echo '{GlobalConfig.PASSWORD}' | sudo -S rm -rf {GlobalConfig.TMPDIR}",
            interrupt=False,
            out_debug_flag=False,
            command_log=False,
        )


# pylint: disable=unused-argument
def pytest_emoji_passed(config, head_line):
    """pytest_emoji_passed"""
    # 笑脸
    return (
        f"【 {datetime.now()} {head_line} || 😃 】\n",
        f"【 {datetime.now()} {head_line} || PASSED 😃 】\n",
    )


# pylint: disable=unused-argument
def pytest_emoji_failed(config, head_line):
    """pytest_emoji_failed"""
    # 哭笑不得
    return (
        f"【 {datetime.now()} {head_line} || 😰 】\n",
        f"【 {datetime.now()} {head_line} || FAILED 😰 】\n",
    )


# pylint: disable=unused-argument
def pytest_emoji_skipped(config, head_line):
    """pytest_emoji_skipped"""
    # 翻白眼儿
    return (
        f"【 {datetime.now()} {head_line} || 🙄 】\n",
        f"【 {datetime.now()} {head_line} || SKIPPED 🙄 】\n",
    )


# pylint: disable=unused-argument
def pytest_emoji_error(config, head_line):
    """pytest_emoji_error"""
    # 哭哭
    return (
        f"【 {datetime.now()} {head_line} || 😡 】\n",
        f"【 {datetime.now()} {head_line} || ERROR 😡 】\n",
    )


class DuringfailingTerminalReporter(TerminalReporter):
    """测试过程中立即显示报错"""

    def __init__(self, reporter):
        TerminalReporter.__init__(self, reporter)
        self._tw = reporter._tw

    def pytest_collectreport(self, report):
        """pytest_collectreport"""
        # 立即显示收集过程中发生的错误。
        TerminalReporter.pytest_collectreport(self, report)
        if report.failed:
            if self.isatty:
                self.rewrite("")
            self.print_failure(report)

    def pytest_runtest_logreport(self, report):
        """pytest_runtest_logreport"""
        # 立刻显示运行测试期间发生的故障和错误
        TerminalReporter.pytest_runtest_logreport(self, report)
        if report.failed and not hasattr(report, "wasxfail"):
            if self.verbosity <= 0:
                self._tw.line()
            self.print_failure(report)

    def summary_failures(self):
        """summary failures"""
        # 防止显示错误摘要，因为我们已经错误发生后立即显示错误。

    def summary_errors(self):
        """summary errors"""

    def print_failure(self, report):
        """print failure"""
        if self.config.option.tbstyle != "no":
            if self.config.option.tbstyle == "line":
                line = self._getcrashline(report)
                self.write_line(line)
            else:
                msg = self._getfailureheadline(report)
                if report.when == "collect":
                    msg = "ERROR collecting " + msg
                elif report.when == "setup":
                    msg = "ERROR at setup of " + msg
                elif report.when == "teardown":
                    msg = "ERROR at teardown of " + msg
                self.write_sep("_", msg)
                if not self.config.getvalue("usepdb"):
                    self._outrep_summary(report)


def get_runs_id_deque(user, password, class_obj, func, _id):
    """get runs id deque"""
    if not (user and password):
        raise ValueError("缺少PMS用户名或密码")
    runs_ids = getattr(class_obj(user, password), f"get_{func}_data")(_id)
    if not runs_ids:
        raise ValueError
    id_deque = deque()
    for i in runs_ids:
        id_deque.append(i.get("case_id"))
        id_deque.append(i.get("from_case_id"))
    return runs_ids, id_deque


def add_run_case_id(session, item, tag, runs_ids):
    """add run case id"""
    if auto_send(session):
        # 需要回填数据的时候才做
        for i in runs_ids:
            _case_id = i.get("case_id")
            _from_case_id = i.get("from_case_id")
            _run_case_id = i.get("run_case_id")
            if tag in (_case_id, _from_case_id):
                add_mark(item, _run_case_id, ("run_case_id",), {})
                add_mark(item, _from_case_id, ("from_case_id",), {})
                break


@pytest.fixture
def __pytest_repeat_step_number(request):
    """pytest repeat step number"""
    marker = request.node.get_closest_marker("repeat")
    repeat = marker and marker.args[0] or request.config.option.repeat
    if repeat > 1:
        return request.param
    return None


def walk_apps(walk_dir):
    """walk apps"""
    no_youqu_mark = {}
    csv_path_dict = {}
    for root, _, files in walk(walk_dir):
        if "NOYOUQUMARK" in files and not no_youqu_mark.get(root):
            no_youqu_mark[root] = True
            continue
        for file in files:
            if file.endswith(".csv") and file != "case_list.csv":
                csv_path_dict[splitext(file)[0]] = f"{root}/{file}"
    return csv_path_dict, no_youqu_mark
